/* 
 * 	Copyright (c) 2010 Colorado State University
 * 
 *	Permission is hereby granted, free of charge, to any person
 *	obtaining a copy of this software and associated documentation
 *	files (the "Software"), to deal in the Software without
 *	restriction, including without limitation the rights to use,
 *	copy, modify, merge, publish, distribute, sublicense, and/or
 *	sell copies of the Software, and to permit persons to whom
 *	the Software is furnished to do so, subject to the following
 *	conditions:
 *
 *	The above copyright notice and this permission notice shall be
 *	included in all copies or substantial portions of the Software.
 *
 *	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *	EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 *	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *	NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *	HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *	FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 *	OTHER DEALINGS IN THE SOFTWARE.\
 * 
 * 
 *  File: mrtinstance.c
 * 	Authors: He Yan, Dan Massey
 *
 *  Date: Oct 7, 2008 
 */
 
/* 
 * Maintain a connection to a mrt box
 */

/* externally visible structures and functions for mrts */
#include "mrt.h"
/* internal structures of Mrt Control module */
#include "mrtcontrol.h"
/* internal structures and functions for this module */
#include "mrtinstance.h"

/* required for logging functions */
#include "../Util/log.h"

/* needed for address management  */
#include "../Util/address.h"

/* needed for writen function  */
#include "../Util/unp.h"

#include "../Util/bgpmon_formats.h"

/* needed for function getXMLMessageLen */
#include "../XML/xml.h"

/* needed for session related function */
#include "../Peering/peersession.h"

/* needed for Mrt Table Dump v2 and state of connection */
#include "../Peering/bgppacket.h"
#include "../Peering/bgpmessagetypes.h"

/* needed for deleteRib table if socket connection lost */
#include "../Labeling/label.h"

/* needed for session State */
#include "../Peering/bgpstates.h"



/* needed for malloc and free */
#include <stdlib.h>
/* needed for strncpy */
#include <string.h>
/* needed for system error codes */
#include <errno.h>
/* needed for addrinfo struct */
#include <netdb.h>
/* needed for system types such as time_t */
#include <sys/types.h>
/* needed for time function */
#include <time.h>
/* needed for socket operations */
#include <sys/socket.h>
/* needed for pthread related functions */
#include <pthread.h>




#include <stdio.h>
#include <arpa/inet.h>
#include <inttypes.h>
#include <netinet/in.h>

//#define DEBUG


/*--------------------------------------------------------------------------------------
 * Purpose: Get an array of IDs of all active mrt connections
 * Input: a pointer to an unallocated array
 * Output: the length of the array or -1 on failure
 * Note: 1. The caller doesn't need to allocate memory.
 *       2. The caller must free the array after using it.
 * He Yan @ Sep 22, 2008
 * -------------------------------------------------------------------------------------*/ 
int 
getActiveMrtsIDs(long **mrtIDs)
{
	// lock the mrt list
	if ( pthread_mutex_lock( &(MrtControls.mrtLock) ) )
		log_fatal("lock mrt list failed");


	// allocate an array whose size depends on the active mrt
	long *IDs = malloc(sizeof(long)*MrtControls.activeMrts);
	if (IDs == NULL) {
		log_err("Failed to allocate memory for getActiveMrtsIDs");
		*mrtIDs = NULL; 
		return -1;
	}


	// for each active mrt, add its ID to the array
	MrtNode *cn = MrtControls.firstNode;
	int i = 0;
	while( cn != NULL )
	{
		IDs[i] = cn->id;
		i++;
		cn = cn->next;
	}
	*mrtIDs = IDs; 

	//sanity check how many IDs we found
	if (i != MrtControls.activeMrts)
	{
		log_err("Unable to get Active Mrt IDs!");
		free(IDs);
		*mrtIDs = NULL; 
		i = -1;
	}

	// unlock the mrt list
	if ( pthread_mutex_unlock( &(MrtControls.mrtLock) ) )
		log_fatal( "unlock mrt list failed");

	return i;
}

/*--------------------------------------------------------------------------------------
 * Purpose: Get the connected mrt's port
 * Input: ID of the mrt
 * Output: port of this mrt 
 *         or -1 if there is no mrt with this ID 
 * He Yan @ Sep 22, 2008
 * -------------------------------------------------------------------------------------*/ 
int 
getMrtPort(long ID)
{
	MrtNode *cn = MrtControls.firstNode;
	while( cn != NULL )
	{
		if(cn->id == ID)
			return cn->port;
		else	
			cn = cn->next;
	}

	log_err("getMrtPort: couldn't find a mrt connection with ID: %d", ID);
	return -1;
}

/*--------------------------------------------------------------------------------------
 * Purpose: Get the connected mrt's address
 * Input: ID of the mrt
 * Output: address of this mrt in a char array. 
 *         or NULL if there is no mrt with this ID 
 * Note: 1. The caller doesn't need to allocate memory.
 *       2. The caller must free the string after using it.
 * He Yan @ Sep 22, 2008
 * -------------------------------------------------------------------------------------*/ 
char * 
getMrtAddress(long ID)
{
	MrtNode *cn = MrtControls.firstNode;
	while( cn != NULL )
	{
		if(cn->id == ID)
		{
			char *ans = malloc( strlen(cn->addr) + 1 );
			if (ans == NULL) {
				log_err("getMrtAddress: couldn't allocate string memory");
				return NULL;
			}
			strncpy(ans, cn->addr, strlen(cn->addr) + 1);
			return ans;
		}
		else	
			cn = cn->next;
	}

	log_err("getMrtAddress: couldn't find a mrt with ID: %d", ID);
	return NULL;
}

/*--------------------------------------------------------------------------------------
 * Purpose: Get the connected time of a mrt
 * Input: ID of the mrt
 * Output: connected time of this mrt
 * 		  or time = 0 if there is no mrt with this ID
 * He Yan @ Sep 22, 2008
 * -------------------------------------------------------------------------------------*/ 
time_t 
getMrtConnectedTime(long ID)
{
	time_t ans = 0;
	MrtNode *cn = MrtControls.firstNode;
	while( cn != NULL )
	{
		if(cn->id == ID)
			return cn->connectedTime;
		else	
			cn = cn->next;
	}

	log_err("getMrtConnectedTime: couldn't find a mrt with ID: %d", ID);
	return ans;
}

/*--------------------------------------------------------------------------------------
 * Purpose: Get the labelAction of a mrt
 * Input: ID of the mrt
 * Output: label Action of this mrt
 * Mikhail Strizhov @ Jan 30, 2009
 * -------------------------------------------------------------------------------------*/ 
int 
getMrtLabelAction(long ID)
{
	int ans = 0;
	MrtNode *cn = MrtControls.firstNode;
	while( cn != NULL )
	{
		if(cn->id == ID)
			return cn->labelAction;
		else	
			cn = cn->next;
	}

	log_err("getMrtConnectedTime: couldn't find a mrt with ID: %d", ID);
	return ans;
}


/*--------------------------------------------------------------------------------------
 * Purpose: Get the number of items written by this mrt connection
 * Input: ID of the mrt
 * Output: the number of items
 * 	   or -1 if there is no mrt with this ID
 * He Yan @ Sep 22, 2008
 * -------------------------------------------------------------------------------------
 */ 
long
getMrtWriteItems(long ID)
{
	MrtNode *cn = MrtControls.firstNode;
	while( cn != NULL )
	{
		if(cn->id == ID)
		{
		}
		else	
			cn = cn->next;
	}

	log_warning("getMrtWriteItems: couldn't find a mrt with ID:%d", ID);
	return -1;
}

/*--------------------------------------------------------------------------------------
 * Purpose: Get the last action time for this mrt connection
 * Input: ID of the mrt
 * Output: a timevalue indicating the last time the thread was active
 *         or time = 0 if there is no mrt with this ID
 * He Yan @ Sep 22, 2008
 * -------------------------------------------------------------------------------------*/
time_t
getMrtLastAction(long ID)
{
	time_t ans = 0;
	MrtNode *cn = MrtControls.firstNode;
	while( cn != NULL )
	{
		if(cn->id == ID)
			return  cn->lastAction;
		else	
			cn = cn->next;
	}

	log_err("getMrtLastAction: couldn't find a mrt with ID: %d", ID);
	return ans;
}

/*--------------------------------------------------------------------------------------
 * Purpose: kill a mrt connection, close the thread and free all associated memory 
 * Input: ID of the mrt
 * He Yan @ Sep 22, 2008
 * -------------------------------------------------------------------------------------*/ 
void 
deleteMrt(long ID)
{
	MrtNode *cn = MrtControls.firstNode;
	log_msg("deleteMrt Called!");
	while( cn != NULL )
	{
		if(cn->id == ID)
		{
			cn->deleteMrt = TRUE;
			return;
		}
		else	
			cn = cn->next;
	}

	log_err("deleteMrt: couldn't find a mrt with ID:%d", ID);
}

/*--------------------------------------------------------------------------------------
 * Purpose:  destroy the socket and memory associated with a mrt ID
 * Input:  the mrt ID to destroy
 * Output: none
 * He Yan @ July 22, 2008
 * -------------------------------------------------------------------------------------*/
void 
destroyMrt(long id)
{
	// lock the mrt list
	if ( pthread_mutex_lock( &(MrtControls.mrtLock) ) )
		log_fatal( "lock mrt list failed");

	MrtNode *prev = NULL;
	MrtNode *cn = MrtControls.firstNode;

	while(cn != NULL) {
		if (cn->id == id ) 
		{
			// close this mrt connection
			log_msg("Deleting mrt id (%d)", cn->id);		
			// close the mrt socket
			close( cn->socket );
			// remove from the mrt list
			MrtControls.activeMrts--;
			if (prev == NULL) 
				MrtControls.firstNode = cn->next;
			else
				prev->next = cn->next;
			// clean up the memory
			destroyQueueWriter( cn->qWriter );
			free(cn);
			// unlock the mrt list
			if ( pthread_mutex_unlock( &(MrtControls.mrtLock) ) )
				log_fatal( "unlock mrt list failed");			
			return;
		}
		else
		{
			prev = cn;
			cn = cn->next;
		}
	}

	log_err("destroyMrt: couldn't find a mrt with ID:%d", id);
	// unlock the mrt list
	if ( pthread_mutex_unlock( &(MrtControls.mrtLock) ) )
		log_fatal( "unlock mrt list failed");
	return;
}

/*--------------------------------------------------------------------------------------
 * Purpose: Create a new MrtNode structure.
 * Input:  the mrt ID, address (as string), port, and socket
 * Output: a pointer to the new MrtNode structure 
 *         or NULL if an error occurred.
 * He Yan @ July 22, 2008
 * -------------------------------------------------------------------------------------*/
MrtNode * 
createMrtNode( long ID, char *addr, int port, int socket, int labelAction )
{
	// create a mrt node structure
	MrtNode *cn = malloc(sizeof(MrtNode));
	if (cn == NULL) {
		log_warning("Failed to allocate memory for new mrt.");
		return NULL;
	}
	cn->id = ID;
	strncpy( cn->addr, addr, strlen(addr) );
	cn->port = port;
	cn->socket = socket;
	cn->connectedTime = time(NULL);
	cn->lastAction = time(NULL);
	cn->qWriter= createQueueWriter( peerQueue );
	cn->deleteMrt = FALSE;		
	cn->next = NULL;
	cn->labelAction = labelAction;
	return cn;
}

/*--------------------------------------------------------------------------------------
 * Purpose: The main function of a thread handling one mrt connection
 * Input:  the mrt node structure for this mrt
 * Output: none
 * He Yan @ July 22, 2008
 * -------------------------------------------------------------------------------------*/
void *
mrtThread( void *arg )
{
	MrtNode *cn = arg;	// the mrt node structure
			
	// detach the thread so the resources may be returned when the thread exits
	pthread_detach(pthread_self());

	// get the xml queue reader
	QueueWriter peerQueueWriter = cn->qWriter;

	time_t		pretime = 0;
	u_int16_t	pretype = 0 ;  
	u_int16_t	presubtype = 0;
	u_int32_t	prelength = 0;
	int 		preremainingLen = 0;
	BMF 		prebmf = NULL;

	int sessionID;
	
	// Added by Mikhail Strizhov  
	// purpose - handle Table Dump v2 messages
	int NumberOfPeers = 0;	

	int i,j;
	char PeerIPString[256];
	char SrcIPString[256];
	int	TD_sessionID[MAX_NUM_PEERS];
	int bytes = 0;
	char str[256];

	int	BMFremaining = 0;	
	int PrefixLenbytes = 0;
		
	int AttrLenCount;
	int BGP_MP_LEN = 5;

	int SessionCounter[MAX_NUM_PEERS];	
	memset(SessionCounter, 0, sizeof(SessionCounter));

	int sessionNumb = 0;
	
	int ASNumLen = 0;
	// while the mrt connection is alive, read MRT data from it,
	// convert it to BMF format and write it into peer queue
	while ( cn->deleteMrt== FALSE )
	{

		// update the last action time
		cn->lastAction = time(NULL);

		// first read the MRT header
		time_t		time;
		u_int16_t	type;
		u_int16_t	subtype;
		u_int32_t	length;
		int 		n; 
		n = readn( cn->socket, &time, 4);
		if( n != 4 )
		{
			log_err("mrtThread, read MRT header: time error!");
			break;
		}

		n = readn( cn->socket, &type, 2);
		if( n != 2 )
		{
			log_err("mrtThread, read MRT header: type error!");
			break;
		}
		
		n = readn( cn->socket, &subtype, 2);
		if( n != 2 )
		{
			log_err("mrtThread, read MRT header: subtype error!");
			break;
		}
		
		n = readn( cn->socket, &length, 4);
		if( n != 4 )
		{
			log_err("mrtThread, read MRT header: length error!");
			break;
		}

		type = ntohs(type);
		subtype = ntohs(subtype);
		time = ntohl(time);
		length = ntohl(length);

	
		// only prccess BGP4MP messages
		if( (type != 16) && (type != 13) )
		{
			log_warning("mrtThread, receive a corrupted MRT message with time %d type %u, subtype %u and len %u!", time, type, subtype, length);
			log_warning("mrtThread, previous messgae with time %d type %u subtype %u, len %u and BGP length %d!", pretype, presubtype, prelength, preremainingLen);
			// read the remaining stuff
			/*
			u_int8_t *remaining = malloc(length);
			n = readn( cn->socket, remaining, length);
			if( n != length )
			{
				log_err("mrtThread, read BGP4MP_MESSAGE message: read remaining stuff error!");
				break;
			}
			free(remaining);
			*/

			// MRT corrupted, find the next valid BGP message
			while(1)
			{
				u_int8_t nextByte = 0;
				int i;
				n = readn( cn->socket, &nextByte, 1);
				if(nextByte == 255)
				{	
					for(i=0; i<15; i++)
					{
						n = readn( cn->socket, &nextByte, 1);
						if( nextByte != 255 )
						{
							break;
						}
					}
					if( i == 15 )
					{
						u_int16_t len;
						u_int8_t *tmpBuf = NULL;
						n = readn( cn->socket, &len, 2);
						len = ntohs(len);
						tmpBuf = malloc(len-18);
						n = readn( cn->socket, tmpBuf, len-18);
						log_msg("mrtThread, find next BGP message with len %d", len);
						free(tmpBuf);
						break;
					}
				}
			}
#ifdef DEBUG
			hexdump(LOG_WARNING, prebmf->message, prebmf->length);
#endif
			destroyBMF(prebmf);
			prebmf = NULL;
			continue;
		}


		// Added by Mikhail Strizhov - Aug 21, 2009		
		// Do we got a mrt TABLE DUMP v2 ?
		if (type == 13)
		{
	
			if (subtype == 1)
			{

				mrt_index *indexPtr = malloc(sizeof(mrt_index));
				if (indexPtr == NULL)
				{
					log_warning("Failed to allocate memory for MRT IPv4 MESSAGE.");
					return NULL;
				}		
		

				n = readn( cn->socket, &indexPtr->BGPSrcID, sizeof(indexPtr->BGPSrcID));
				if( n != sizeof(indexPtr->BGPSrcID) )
				{
					log_err("mrtThread, read MRT header: indexPtr->BGPSrcID error!");
					break;
				}
	
				if( inet_ntop(AF_INET, &indexPtr->BGPSrcID, SrcIPString, ADDR_MAX_CHARS) == NULL )
				{
					log_err("mrtThread, TABLE DUMP v2 message: indexPtr->BGPSrcID ipv4 convert error!");
					break;
				}
		

				n = readn( cn->socket, &indexPtr->ViewNameLen, sizeof(indexPtr->ViewNameLen));
				if( n != sizeof(indexPtr->ViewNameLen) )
				{
					log_err("mrtThread, read MRT header: indexPtr->ViewNameLen error!");
					break;
				}

				if (indexPtr->ViewNameLen != 0)
				{
					n = readn( cn->socket, &indexPtr->ViewName, sizeof (indexPtr->ViewName));
					if( n != sizeof (indexPtr->ViewName) )
					{
						log_err("mrtThread, read MRT header: indexPtr->ViewName error!");
						break;
					}
				}

				n = readn( cn->socket, &indexPtr->PeerCount, sizeof (indexPtr->PeerCount));
				if( n != sizeof (indexPtr->PeerCount) )
				{
					log_err("mrtThread, read MRT header: indexPtr->PeerCount error!");
					break;
				}
				indexPtr->PeerCount = ntohs (indexPtr->PeerCount);
				NumberOfPeers = indexPtr->PeerCount;
				for (i = 0; i < (indexPtr->PeerCount); i++)
				{
					// read the PeerType
					n = readn( cn->socket, &indexPtr->PeerType, sizeof (indexPtr->PeerType));
					if( n != sizeof (indexPtr->PeerType) )
					{
						log_err("mrtThread, read MRT header: indexPtr->PeerType error!");
						break;
					}
			
					// read the BGP ID 
					n = readn( cn->socket, &indexPtr->PeerBGPID, sizeof (indexPtr->PeerBGPID));
					if( n != sizeof (indexPtr->PeerBGPID) )
					{
						log_err("mrtThread, read MRT header: indexPtr->PeerBGPID error!");
						break;
					}			
					// check 1st bit - IPv4 or IPv6   
					if ( (indexPtr->PeerType & 0x01) == 0) 
					{
						n = readn( cn->socket, &indexPtr->PeerIP, 4);
						if( n != 4 )
						{
							log_err("mrtThread, read MRT header: indexPtr->PeerIP IPv4 error!");
							break;
						}
				
						if( inet_ntop(AF_INET, &indexPtr->PeerIP, PeerIPString, ADDR_MAX_CHARS) == NULL )
						{
							log_err("mrtThread, TABLE DUMP v2 message: ipv4 source address convert error!");
							break;
						}
					}
					if ( (indexPtr->PeerType & 0x01) == 1)
					{
						n = readn( cn->socket, &indexPtr->PeerIP, 16);
						if( n != 16 )
						{
							log_err("mrtThread, read MRT header: indexPtr->PeerIP IPv6 error!");
							break;
						}
						if( inet_ntop(AF_INET6, &indexPtr->PeerIP, PeerIPString, ADDR_MAX_CHARS) == NULL )
						{
							log_err("mrtThread, TABLE DUMP v2 message: ipv4 source address convert error!");
							break;
						}				
					}
					// check 2nd bit - ASN2 or ASN4
					if ( (indexPtr->PeerType & 0x02) == 0)
					{
						// set AS Num Len to 2 byte
						ASNumLen = 2;
						// read from socket PeerAS
						n = readn( cn->socket, &indexPtr->PeerAS, 2);
						if( n != 2 )
						{
							log_err("mrtThread, read MRT header: indexPtr->PeerAS error! with %d ASNumLen\n", ASNumLen);
							break;
						}
						indexPtr->PeerAS = ntohs (indexPtr->PeerAS);
#ifdef DEBUG						
						printf("\n Peer AS is %d \n", indexPtr->PeerAS);
#endif						
					}
					if ( (indexPtr->PeerType & 0x02) == 2)
					{
						// set AS Num Len to 4 byte
						ASNumLen = 4;
						// read from socket PeerAS
						n = readn( cn->socket, &indexPtr->PeerAS, 4);
						if( n != 4 )
						{
							log_err("mrtThread, read MRT header: indexPtr->PeerAS error! with %d ASNumLen\n", ASNumLen);
							break;
						}
						indexPtr->PeerAS = ntohl (indexPtr->PeerAS);
#ifdef DEBUG						
						printf("\n Peer AS is %d \n", indexPtr->PeerAS);
#endif						
					}
					
					if (ASNumLen == 0)
					{
						log_warning("mrtThread, ASNumLen is not 2 nor 4 bytes!\n");
						break;
					}
					
			  		// find a session based on six tuples
					sessionID = findSession(  indexPtr->PeerAS, 6447, 179, 179, PeerIPString, SrcIPString);
					if( sessionID < 0)
					{
						// set mrt table flag to 1
						int mrtTableFlag = 1;
						// create a new session
						TD_sessionID[i] = createMrtSessionStruct(  indexPtr->PeerAS, 6447, PeerIPString, SrcIPString, cn->labelAction, ASNumLen, mrtTableFlag);
						if( TD_sessionID[i] < 0 )
						{
							log_err( "mrtThread, fail to create a new session for AS 6447,AS %d, SRC %s, DST %s",   indexPtr->PeerAS, SrcIPString, PeerIPString );
							break;
						}						
						else
						{	
							log_msg( "mrtThread, create a new session for AS 6447, AS %d, SRC %s, DST %s, ASNumLen %d",  indexPtr->PeerAS, SrcIPString, PeerIPString, ASNumLen );
						}
					}
					else
					{
						// store all SessionID's in array, we will need it in new BMF messages
						TD_sessionID[i] = sessionID;
					}
				
					// if we 4 byte ASN updates and 2 byte tables,   not possible!!
					if ((getSessionMrtTableFlag(TD_sessionID[i]) == 0) && (getSessionASNumberLength(TD_sessionID[i]) != ASNumLen))
					{
						if (ASNumLen == 2)
							log_err("mrtThread, Session %d Updates are %d byte ASNs,  but RIB ASNumLen is %d bytes",TD_sessionID[i], getSessionASNumberLength(TD_sessionID[i]), ASNumLen) ;
					}

				} // end of  peer number loop
				free(indexPtr);
			} // end of subtype == 1
	
			// type 2 - RIB_IPV4_UNICAST subtype
			if (subtype == 2)
			{
				if (NumberOfPeers == 0)
				{
					log_err("mrtThread, No NumberOfPeers in IPv4 subtype 2");
					break;
				}

				mrt_uni *uniPtr = malloc(sizeof(mrt_uni));
				if ( uniPtr == NULL) {
					log_warning("Failed to allocate memory for MRT IPv4 MESSAGE.");
					return NULL;
				}
		
				n = readn( cn->socket, &uniPtr->SeqNumb, sizeof(uniPtr->SeqNumb));
				if( n != sizeof(uniPtr->SeqNumb) )
				{
					log_err("mrtThread, read MRT header: uniPtr->SeqNumb error!");
					break;
				}	

				n = readn( cn->socket, &uniPtr->PrefixLen, sizeof(uniPtr->PrefixLen));
				if( n != sizeof(uniPtr->PrefixLen) )
				{
					log_err("mrtThread, read MRT header: uniPtr->PrefixLen error!");
					break;
				}	

				memset(uniPtr->Prefix, 0, sizeof(uniPtr->Prefix));
		
				if (uniPtr->PrefixLen != 0)
				{
					bytes = 0;
					for (i = 0; i < 128; i=i+8)
					{
						bytes++;
						if ( (uniPtr->PrefixLen > i) && (uniPtr->PrefixLen <= i+8) )
						{	
							n = readn( cn->socket, (uniPtr->Prefix), bytes);
							if( n != bytes )
							{
								log_err("mrtThread, read MRT header: uniPtr->Prefix error!");
								break;
							}					
#ifdef DEBUG							
							inet_ntop(AF_INET, (uniPtr->Prefix), str, ADDR_MAX_CHARS);
							printf("RIP IPV4 UNICAST, Peer's Prefix is  %s / %d, i is %d, bytes are %d\n", str,uniPtr->PrefixLen,  i, bytes);									
#endif		
							break;
						}

					}
			
				}
				else
				{
					inet_ntop(AF_INET, (uniPtr->Prefix), str, ADDR_MAX_CHARS);
					log_msg("ZERO IPV4 UNICAST, Peer's Prefix is  %s", str);				
				}

				n = readn( cn->socket, &uniPtr->EntryCount, sizeof(uniPtr->EntryCount));
				if( n != sizeof(uniPtr->EntryCount) )
				{
					log_err("mrtThread, read MRT header: uniPtr->EntryCount error!");
					break;
				}	
				uniPtr->EntryCount = ntohs (uniPtr->EntryCount);
				for (j = 0; j < (uniPtr->EntryCount); j++)
				{
					n = readn( cn->socket, &uniPtr->PeerIndex,  sizeof(uniPtr->PeerIndex));
					if( n != sizeof(uniPtr->PeerIndex) )
					{
						log_err("mrtThread, read MRT header: uniPtr->PeerIndex error!");
						break;
					}				
				   	uniPtr->PeerIndex = ntohs(uniPtr->PeerIndex);
		
					n = readn( cn->socket, &uniPtr->OrigTime, sizeof(uniPtr->OrigTime));
					if( n != sizeof(uniPtr->OrigTime) )
					{
						log_err("mrtThread, read MRT header: uniPtr->OrigTime error!");
						break;
					}				
					uniPtr->OrigTime = ntohl(uniPtr->OrigTime);
			
					n = readn( cn->socket, &uniPtr->AttrLen, sizeof(uniPtr->AttrLen));
					if( n != sizeof(uniPtr->AttrLen) )
					{
						log_err("mrtThread, read MRT header: uniPtr->AttrLen error!");
						break;
					}				
					uniPtr->AttrLen = ntohs(uniPtr->AttrLen);
					AttrLenCount = uniPtr->AttrLen;


					n = readn( cn->socket, uniPtr->BGPAttr, uniPtr->AttrLen);
					if( n != uniPtr->AttrLen )
					{
						log_err("mrtThread, read MRT header: uniPtr->BGPAttr error!");
						break;
					}	

					bytes = 0;
					for (i = 0; i < 128; i=i+8)
					{
						bytes++;
						if ( (uniPtr->PrefixLen > i) && (uniPtr->PrefixLen <= i+8) )
						{	

							PrefixLenbytes = bytes;
//							printf("PrefixLenbytes is %d \n", PrefixLenbytes);
							break;
						}
					}
					// 5 is the size of WDLen, PrefixLen, AttrLen
					BMFremaining = uniPtr->AttrLen + PrefixLenbytes + 5;
		
					// Creating BGP Message 
					mrt_ipv4_message *ipv4Ptr = malloc(sizeof(mrt_ipv4_message));
					if ( ipv4Ptr == NULL) {
						log_warning("Failed to allocate memory for MRT IPv4 MESSAGE.");
						return NULL;
					}
			
					ipv4Ptr->WDLen = htons(0);
					ipv4Ptr->AttrLen = htons(uniPtr->AttrLen);
					memset(ipv4Ptr->mainmsg, 0, sizeof(ipv4Ptr->mainmsg));
					memcpy(ipv4Ptr->mainmsg, uniPtr->BGPAttr, AttrLenCount);
					memcpy(&ipv4Ptr->mainmsg[AttrLenCount], &uniPtr->PrefixLen, 1);
					memcpy(&ipv4Ptr->mainmsg[AttrLenCount+1], uniPtr->Prefix, PrefixLenbytes);				

					for (i = 0; i < NumberOfPeers; i++)
					{	
						if (uniPtr->PeerIndex == i)
						{	
					
						BMF bmf = NULL;
						bmf = createBMF(TD_sessionID[i],  BMF_TYPE_TABLE_TRANSFER);
						memcpy(&bmf->timestamp, &uniPtr->OrigTime, sizeof(uniPtr->OrigTime));
						memset(&bmf->precisiontime, 0, sizeof(bmf->precisiontime));
		
						PBgpHeader hdr = NULL; 
						hdr = createBGPHeader( typeUpdate ); 
						hdr->length += htons(BMFremaining);
				
						bgpmonMessageAppend( bmf, hdr, 19);	
						bgpmonMessageAppend( bmf, ipv4Ptr, BMFremaining);
						writeQueue( peerQueueWriter, bmf );
						incrementSessionMsgCount(TD_sessionID[i]);
						free(hdr);
						break;	
						}			
					}
					free(ipv4Ptr);
				}
				free(uniPtr);
			} // end of subtype == 2



			// type 4 - RIB_IPV6_UNICAST subtype
			if (subtype == 4)
			{
				if (NumberOfPeers == 0)
				{
					printf("mrtThread, No NumberOfPeers in IPv6 subtype 4");
					break;
				}
		
				mrt_uni *uniPtr = malloc(sizeof(mrt_uni));
				if ( uniPtr == NULL) {
					log_warning("Failed to allocate memory for MRT IPv6 MESSAGE.");
					return NULL;
				}

				n = readn( cn->socket, &uniPtr->SeqNumb, sizeof(uniPtr->SeqNumb) );
				if (n != sizeof(uniPtr->SeqNumb))
				{
					log_err("mrtThread, read MRT header: uniPtr->SeqNumb ipv6 error!");
					break;
				}

				n = readn( cn->socket, &uniPtr->PrefixLen, sizeof(uniPtr->PrefixLen) );
				if (n != sizeof(uniPtr->PrefixLen))
				{
					log_err("mrtThread, read MRT header: uniPtr->PrefixLen ipv6 error!");
					break;
				}

				memset(uniPtr->Prefix, 0, sizeof(uniPtr->Prefix));
				if (uniPtr->PrefixLen != 0)
				{
					bytes = 0;
					for (i = 0; i <= 128; i=i+8)
					{
						bytes++;
						if ((uniPtr->PrefixLen > i) && (uniPtr->PrefixLen) <= i+8 )
						{	
							n = readn( cn->socket, uniPtr->Prefix, bytes );
							if (n != bytes)
							{
								log_err("mrtThread, read MRT header: uniPtr->Prefix ipv6 error!");
								break;
							}
#ifdef DEBUG							
							inet_ntop(AF_INET6, &uniPtr->Prefix, str, ADDR_MAX_CHARS);
							printf("RIP IPV6 UNICAST, Peer's Prefix is   %s / %d, i is %d, bytes are %d\n", str,uniPtr->PrefixLen, i, bytes);
#endif
							break;		
						}
					}
			
				}
				else
				{
					inet_ntop(AF_INET6, &(uniPtr->Prefix), str, ADDR_MAX_CHARS);
					log_msg("ZERO IPV6 UNICAST, Peer's Prefix is  %s\n", str);				
				}

				n = readn( cn->socket, &uniPtr->EntryCount, sizeof(uniPtr->EntryCount) );
				if (n != sizeof(uniPtr->EntryCount))
				{
					log_err("mrtThread, read MRT header: uniPtr->EntryCount ipv6 error!");
					break;
				}
		
				uniPtr->EntryCount = ntohs (uniPtr->EntryCount);
				for (j = 0; j < uniPtr->EntryCount; j++)
				{
			
					n = readn( cn->socket, &uniPtr->PeerIndex,  sizeof(uniPtr->PeerIndex) );
					if (n != sizeof(uniPtr->PeerIndex))
					{
						log_err("mrtThread, read MRT header: uniPtr->PeerIndex ipv6 error!");
						break;
					}			
				   	uniPtr->PeerIndex = ntohs(uniPtr->PeerIndex);
		
					n = readn( cn->socket, &uniPtr->OrigTime, sizeof(uniPtr->OrigTime) );
					if (n != sizeof(uniPtr->OrigTime))
					{
						log_err("mrtThread, read MRT header: uniPtr->OrigTime ipv6 error!");
						break;
					}				
					uniPtr->OrigTime = ntohl(uniPtr->OrigTime);
			
					n = readn( cn->socket, &uniPtr->AttrLen, sizeof (uniPtr->AttrLen) );
					if (n != sizeof(uniPtr->AttrLen))
					{
						log_err("mrtThread, read MRT header: uniPtr->AttrLen ipv6 error!");
						break;
					}			
					uniPtr->AttrLen = ntohs(uniPtr->AttrLen);
					AttrLenCount = uniPtr->AttrLen;

					n = readn( cn->socket, uniPtr->BGPAttr, uniPtr->AttrLen );
					if (n != uniPtr->AttrLen)
					{
						log_err("mrtThread, read MRT header: uniPtr->BGPAttr ipv6 error!");
						break;
					}	

					bytes = 0;
					for (i = 0; i < 128; i=i+8)
					{
						bytes++;
						if ( (uniPtr->PrefixLen > i) && (uniPtr->PrefixLen <= i+8) )
						{	

							PrefixLenbytes = bytes;
							break;
						}
					}
					// 5 is the size of WDLen, PrefixLen, AttrLen,  BGP_MP_LEN is MP_Reach Len value
					BMFremaining = uniPtr->AttrLen + PrefixLenbytes + BGP_MP_LEN + 5;

					// Creating BGP Message 
					mrt_ipv6_message *ipv6Ptr = malloc(sizeof(mrt_ipv6_message));
					if (ipv6Ptr == NULL) {
						log_warning("Failed to allocate memory for MRT IPv4 MESSAGE.");
						return NULL;
					}
				
					ipv6Ptr->WDLen = htons(0);
					ipv6Ptr->AttrLen = htons(uniPtr->AttrLen);
					memset(ipv6Ptr->mainmsg, 0,sizeof(ipv6Ptr->mainmsg));
					memcpy(ipv6Ptr->mainmsg, uniPtr->BGPAttr, AttrLenCount);
					u_int16_t AFI = 2;
					memcpy(&ipv6Ptr->mainmsg[AttrLenCount], "\0", 1);
					memcpy(&ipv6Ptr->mainmsg[AttrLenCount+1], &AFI , 1);
					u_int8_t SAFI = 1;
					memcpy(&ipv6Ptr->mainmsg[AttrLenCount+2], &SAFI, 1);
					u_int8_t NextHopLen = 0;
					memcpy(&ipv6Ptr->mainmsg[AttrLenCount+3], &NextHopLen, 1);
					u_int8_t SNPANum = 0;
					memcpy(&ipv6Ptr->mainmsg[AttrLenCount+4], &SNPANum, 1);
					memcpy(&ipv6Ptr->mainmsg[AttrLenCount+5], &uniPtr->PrefixLen, 1);
					memcpy(&ipv6Ptr->mainmsg[AttrLenCount+6], uniPtr->Prefix, PrefixLenbytes);

					for (i = 0; i < NumberOfPeers; i++)
					{	
						if (uniPtr->PeerIndex == i)
						{	
						BMF bmf = NULL;
						bmf = createBMF(TD_sessionID[i],  BMF_TYPE_TABLE_TRANSFER);
						memcpy(&bmf->timestamp, &uniPtr->OrigTime, sizeof(uniPtr->OrigTime));
						memset(&bmf->precisiontime, 0, sizeof(bmf->precisiontime));
					
						PBgpHeader hdr = NULL; 
						hdr = createBGPHeader( typeUpdate ); 
						hdr->length += htons(BMFremaining);

						bgpmonMessageAppend( bmf, hdr, 19);	
						bgpmonMessageAppend( bmf, ipv6Ptr, BMFremaining);

						writeQueue( peerQueueWriter, bmf );
						incrementSessionMsgCount(TD_sessionID[i]);
						free(hdr);	
						break;	
						}			
					}
					
					free(ipv6Ptr);			
					 

				}
				free(uniPtr);
			} // end of subtype == 4
		pretime = time;
		pretype = type;
		presubtype = subtype;
		prelength = length;
	
		} // end of if (type == 13)

		if ( type == 16)
		{	
			// only handle BGP4MP_MESSAGE
			if( subtype != 0 && subtype != 1 && subtype != 4 && subtype != 5 )
			{
				log_warning("mrtThread, receive a corrupted MRT message with type %u, subtype %u and len %u!", type, subtype, length);
				log_warning("mrtThread, previous messgae with time %d type %u subtype %u, len %u and BGP length %d!", pretype, presubtype, prelength, preremainingLen);
				// MRT corrupted, find the next valid BGP message
				while(1)
				{
					u_int8_t nextByte = 0;
					int i;
					n = readn( cn->socket, &nextByte, 1);
					if(nextByte == 255)
					{	
						for(i=0; i<15; i++)
						{
							n = readn( cn->socket, &nextByte, 1);
							if( nextByte != 255 )
							{
								break;
							}
						}
						if( i == 15 )
						{
							u_int16_t len;
							u_int8_t *tmpBuf = NULL;
							n = readn( cn->socket, &len, 2);
							len = ntohs(len);
							tmpBuf = malloc(len-18);
							n = readn( cn->socket, tmpBuf, len-18);
							log_msg("mrtThread, find next BGP message with len %d", len);
							free(tmpBuf);
							break;
						}
					}
				}
#ifdef DEBUG				
				hexdump(LOG_WARNING, prebmf->message, prebmf->length);
#endif				
				destroyBMF(prebmf);
		                prebmf = NULL;
				continue;			
			}
			else if( subtype == 0 || subtype == 4)
			{
#ifdef DEBUG
				log_warning("mrtThread, receive a MRT message with type 16 and subtype %d!", subtype);
#endif 				
				// read the remaining stuff
				u_int8_t *remaining = malloc(length);
				n = readn( cn->socket, remaining, length);
				if( n != length )
				{
					log_err("mrtThread, read BGP4MP_MESSAGE message: read remaining stuff error!");
					break;
				}
				free(remaining);			
				continue;
			}
			else if( subtype == 1 || subtype == 5)
			{
				u_int32_t	scrAS;
				u_int32_t	dstAS;
	
				u_int16_t	ifidx;
				u_int16_t	afi;
				u_int8_t	scrAddr[16];
				u_int8_t	dstAddr[16];
				char		scrAddrString[256];
				char		dstAddrString[256];
				int 		remainingLen = 0;

				memset(scrAddr, 0, 16);
				memset(dstAddr, 0, 16);
				memset(scrAddrString, 0, 256);
				memset(dstAddrString, 0, 256);
			
				// BGP4MP_MESSAGE - 2 bytes AS
				if (subtype == 1)
				{
					// set ASNumLen to 2
					ASNumLen = 2;
		
					n = readn( cn->socket, &scrAS, 2);
					if( n != 2 )
					{
						log_err("mrtThread, read BGP4MP_MESSAGE message: source AS error with %d subtype", subtype);
						break;
					}
	
					n = readn( cn->socket, &dstAS, 2);
					if( n != 2 )
					{
						log_err("mrtThread, read BGP4MP_MESSAGE message: destination AS error with %d subtype", subtype);
						break;
					}
					scrAS = ntohs(scrAS);
					dstAS = ntohs(dstAS);
				}
				// BGP4MP_MESSAGE_AS4 - 4 bytes AS
				if (subtype == 5)
				{
					// set ASNumLen to 4
					ASNumLen = 4;

					n = readn( cn->socket, &scrAS, 4);
					if( n != 4 )
					{
						log_err("mrtThread, read BGP4MP_MESSAGE message: source AS error with %d subtype!", subtype);
						break;
					}
	
					n = readn( cn->socket, &dstAS, 4);
					if( n != 4 )
					{
						log_err("mrtThread, read BGP4MP_MESSAGE message: destination AS error with %d subtype!", subtype);
						break;
					}
					scrAS = ntohl(scrAS);
					dstAS = ntohl(dstAS);
				}
				
				
				n = readn( cn->socket, &ifidx, 2);
				if( n != 2 )
				{
					log_err("mrtThread, read BGP4MP_MESSAGE message: interface index error!");
					break;
				}

				n = readn( cn->socket, &afi, 2);
				if( n != 2 )
				{
					log_err("mrtThread, read BGP4MP_MESSAGE message: afi error!");
					break;
				}

				ifidx = ntohs(ifidx);
				afi = ntohs(afi);

				if(afi == 1) //ipv4
				{
					n = readn( cn->socket, scrAddr, 4);
					if( n != 4 )
					{
						log_err("mrtThread, read BGP4MP_MESSAGE message: ipv4 source address error!");
						break;
					}
					if( inet_ntop(AF_INET, scrAddr, scrAddrString, ADDR_MAX_CHARS) == NULL )
					{
						log_err("mrtThread, read BGP4MP_MESSAGE message: ipv4 source address convert error!");
						break;
					}
							
					n = readn( cn->socket, dstAddr, 4);
					if( n != 4 )
					{
						log_err("mrtThread, read BGP4MP_MESSAGE message: ipv4 destination address error!");
						break;
					}
					if( inet_ntop(AF_INET, dstAddr, dstAddrString, ADDR_MAX_CHARS) == NULL )
					{
						log_err("mrtThread, read BGP4MP_MESSAGE message: ipv4 destination address convert error!");
						break;
					}	
					remainingLen = length - 16;
				}
				else if(afi == 2) //ipv6
				{
					n = readn( cn->socket, scrAddr, 16);
					if( n != 16 )
					{
						log_err("mrtThread, read BGP4MP_MESSAGE message: ipv6 source address error!");
						break;
					}
					if( inet_ntop(AF_INET6, scrAddr, scrAddrString, ADDR_MAX_CHARS) == NULL )
					{
						log_err("mrtThread, read BGP4MP_MESSAGE message: ipv6 source address convert error!");
						break;
					}

					n = readn( cn->socket, dstAddr, 16);
					if( n != 16 )
					{
						log_err("mrtThread, read BGP4MP_MESSAGE message: ipv6 destination address error!");
						break;
					}
					if( inet_ntop(AF_INET6, dstAddr, dstAddrString, ADDR_MAX_CHARS) == NULL )
					{
						log_err("mrtThread, read BGP4MP_MESSAGE message: ipv6 destination address convert error!");
						break;
					}		
					remainingLen = length - 40;
				}
				else
				{
					log_err("mrtThread, read BGP4MP_MESSAGE message: afi value error!");
					break;
				}
				
				// check if ASNum is not 0
				if (ASNumLen == 0)
				{
					log_err("mrtThread, ASNumLen is not 2 nor 4 bytes!\n");
					break;
				}

		  		// find a session based on six tuples
				sessionID = findSession( scrAS, dstAS, 179, 179, scrAddrString, dstAddrString );		
				//log_msg( "mrtThread, receive a update from %s to %s", scrAddrString, dstAddrString );
				if( sessionID < 0)
				{
					int mrtTableFlag = 0;
					// create a new session
					sessionID = createMrtSessionStruct( scrAS, dstAS, scrAddrString, dstAddrString, cn->labelAction, ASNumLen, mrtTableFlag);
					if( sessionID < 0 )
					{
						log_err( "mrtThread, fail to create a new session for %d,%d, %s, %s",  scrAS, dstAS, scrAddrString, dstAddrString );
						break;
					}						
					else
					{
						log_msg( "mrtThread, create a new session for %d,%d, %s, %s, ASNumLen %d",  scrAS, dstAS, scrAddrString, dstAddrString, ASNumLen );
						SessionCounter[sessionNumb] = sessionID;
//						printf("SessionCounter is %d\n", SessionCounter[sessionNumb]);
						sessionNumb++;
					}
				}
				//  if updates or tables previously came in 2 bytes,  they cant now switch to 4 bytes
				if ((getSessionASNumberLength(sessionID) == 2 ) && (ASNumLen == 4) )
				{
					log_err("mrtThread, Session %d set its ASNumLen to %d bytes, but received UPDATE ASNumLen %d byte update",sessionID, getSessionASNumberLength(sessionID), ASNumLen) ;
				}
				//  if a previous table message set us to 4 bytes,   reset the ASN size to two bytes
				if ((getSessionASNumberLength(sessionID) == 4 ) && (ASNumLen == 2) && (getSessionMrtTableFlag(sessionID) == 1) )
				{
					log_warning("mrtThread, Session %d set its ASNumLen to %d bytes due to a table message, first update reset the ASNumLen to %d bytes",sessionID, getSessionASNumberLength(sessionID), ASNumLen) ;
					//  set the ASN to 2 and set the Setbytable flag to 0
					if (setSessionASNumberLength(sessionID, 2) !=0)
					{
						log_err("mrtThread, could not set ASNumberLenght");
					}
					if (setSessionMrtTableFlag(sessionID, 0) !=0)
					{
						log_err("mrtThread, could not set mrtTableFlag");
					}
				}


				// read the BGP raw message and make BMF message
				u_int8_t *bgpBuf = malloc(remainingLen);
				n = readn( cn->socket, bgpBuf, remainingLen);
				if( n != remainingLen )
				{
					log_err("mrtThread, read BGP4MP_MESSAGE message: read raw BGP error!");
					break;
				}			
			
				BMF bmf = NULL;
				bmf = createBMF(sessionID,  BMF_TYPE_MSG_FROM_PEER);
				bgpmonMessageAppend( bmf, bgpBuf, remainingLen);
				free(bgpBuf);
			
				if( prebmf != NULL )
				{
					writeQueue( peerQueueWriter, prebmf );
					incrementSessionMsgCount(sessionID);
				}
				prebmf = bmf;

				// read from tcp connection
				/*
				char buf[256];
				int n = readn( cn->socket, buf, 256);
				log_msg("number :%d", n);
				// if read fails, delete the mrt connection
				if ( n != 256 ) 
				{
					cn->deleteMrt= TRUE;
				}
				// otherwise process the mrt data
				else 
				{
					hexdump(LOG_INFO, buf, 256);
				
				}	
				*/
				preremainingLen = remainingLen;

			}

		pretime = time;
		pretype = type;
		presubtype = subtype;
		prelength = length;
	
		}

	}


	// clean all RIB tables with created session id's
/*	if (sessionNumb !=0)
	{
		
		for (i = 0; i < sessionNumb; i++)
		{
			log_warning("Mrt id (%d) connection lost, cleaning RIB table from %d Session", cn->id, SessionCounter[i]);
			if (cleanRibTable(SessionCounter[i]) == -1)
			{
				log_err("mrtThread, cleanRibTable can not with Session %d", SessionCounter[i]);
			}
			changeMrtState( SessionCounter[i], stateError );
		}
	}
*/	

	// destroy the mrt for current connection
	destroyMrt(cn->id);		

	
	pthread_exit( (void *) 1 ); 



}




