/*
 *  Copyright (c) 2010 Colorado State University
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use,
 *  copy, modify, merge, publish, distribute, sublicense, and/or
 *  sell copies of the Software, and to permit persons to whom
 *  the Software is furnished to do so, subject to the following
 *  conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 *  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 *  OTHER DEALINGS IN THE SOFTWARE.
 *
 *
 *  File:    xmldata.c
 *  Authors: Pei-chun Cheng (Modified from xml.c by He Yan)
 *  Date:    Dec 20, 2008
 *
 */

 /*
 * The purpose of xmldata.c is to convert an internal BMF to a xml tree or string.
 * We may further export the xml tree with different formats.
 * Currely, we only export plain xml string.
 */

/* needed for malloc and free */
#include <stdlib.h>

/* needed for function inet_ntoa */
#include <arpa/inet.h>

/* needed for time related functions */
#include <time.h>

/* needed for string and math operation */
#include <string.h>
#include <math.h>
#include <stdio.h>
#include <limits.h>

/* needed for xml operation tring and math operation */
#include <libxml/parser.h>
#include <libxml/tree.h>

/* needed for queues*/
#include "../Queues/queue.h"

/* needed for chains*/
#include "../Chains/chains.h"

/* needed for TRUE/FALSE definitions */
#include "../Util/bgpmon_defaults.h"

/* needed for logging definitions */
#include "../Util/log.h"

/* needed for parsing bgp messages */
#include "../Peering/peersession.h"
#include "../Peering/bgppacket.h"
#include "../Peering/bgpmessagetypes.h"

/* needed for getSessionString */
#include "../Labeling/label.h"

/* needed for GMT_TIME_STAMP and ASCII_MESSAGES flags */
#include "../site_defaults.h"
#include <netinet/in.h>
#include <sys/socket.h>

/* needed for accessing bgpmon listening addr and port */
#include "../Clients/clientscontrol.h"

/* needed for interanl functions */
#include "xmlinternal.h"

#include "xmldata.h"

//#define DEBUG


/* constant */
static char* _VERSION = XFB_VERSION;  /* The implemented XFB version,    specified in xmldata.h */
static char* _XMLNS   = XFB_NS;       /* The implemented XFB name space, specified in xmldata.h */

/* buffer used for xml conversion */
static char _XML[XML_BUFFER_LEN];    /* Global xml buffer, XML_BUFFER_LEN defined in xmlinternal.h */
static int  _XML_LEN_DIGITS = 0;     /* Global variable, number of digits for XML message length,
                                        would be calculated by ceil(log10(XML_BUFFER_LEN)) 
                                      */

/* for improving perfomance */
static xmlNodePtr static_marker_node = NULL;

/* placeholder struct for STATE report data, only for internal use */
typedef struct
{
    long current;
    long avg;
    long sdv;
    long min;
    long max;
    long accu;
    long limit;
} stat_data_t; 

/* placeholder struct for TIME report data, only for internal use */
typedef struct
{
    long current;
    long first_startup;
    long first_down;
    long last_startup;
    long last_down;
    long first_action;
    long last_action;
} time_data_t; 

/*----------------------------------------------------------------------------------------
 * Purpose: add an afi property (attribute) to an xml node
 * input:   node  - pointer to the xml node
 *          afi   - an afi number
 * Output:  the xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
xmlNewPropAFI(xmlNodePtr node, int afi)
{
    char *afi_str   = ""; /* null string */

    switch(afi)
    {
        case BGP_AFI_IPv4:  afi_str = "IPV4";  break;
        case BGP_AFI_IPv6:  afi_str = "IPV6";  break;
        default:            afi_str = "OTHER"; break;
    }

    xmlNewPropString(node, "afi",       afi_str);  /* AFI string, ex: IPV4 (human   readable)*/
    xmlNewPropInt(node,    "afi_value", afi);      /* AFI value,  ex: 1    (machine readable)*/

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: add a safi property (attribute) to an xml node
 * input:   node  - pointer to the xml node
 *          safi  - a  safi number
 * Output:  the xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
xmlNewPropSAFI(xmlNodePtr node, int safi)
{
    char *safi_str   = ""; /* null string */

    switch(safi)
    {
        case BGP_MP_SAFI_UNICAST:    safi_str = "UNICAST";   break;
        case BGP_MP_SAFI_MULTICAST:  safi_str = "MULTICAST"; break;
        case BGP_MP_SAFI_MPLS:       safi_str = "MPLS";      break;
        default: safi_str = "OTHER";     break;
    }

    xmlNewPropString(node,  "safi",       safi_str);  /* SAFI string, ex: UNICAST (human   readable)*/
    xmlNewPropInt(node,     "safi_value", safi);      /* SAFI value,  ex: 1       (machine readable)*/

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate a xml node with a BGPID
 * input:   tag   - tag string
 *          ip    - a long long integer for ip address
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 *
 * Ex:
 *     <BGPID>128.223.51.102</BGPID>
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
xmlNewNodeBGPID(char *tag, long long ip)
{
    static char buff[XML_TEMP_BUFFER_LEN];
    memset(buff, 0, XML_TEMP_BUFFER_LEN);
    buff[0] = '\0';

    if (ip > LONG_MAX) { inet_ntop(AF_INET6, &ip, buff, XML_TEMP_BUFFER_LEN); } /* >  4 bytes */
    else               { inet_ntop(AF_INET,  &ip, buff, XML_TEMP_BUFFER_LEN); } /* <= 4 bytes */

    return xmlNewNodeString(tag, buff);
}

/*----------------------------------------------------------------------------------------
 * Purpose: add a BGPID child node to a xml node
 * input:   parent_node - parent xml node
 *          tag   - tag string for the child node
 *          ip    - a "long long" integer ip for the child node
 * Output:  the child node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
xmlNewChildBGPID(xmlNodePtr parent_node, char *tag, long long ip)
{
    return xmlAddChild(parent_node, xmlNewNodeBGPID(tag, ip));
}


/*----------------------------------------------------------------------------------------
 * Purpose: generate a xml node with general network address, could be v4 or v6
 * input:   tag  - tag string
 *          addr - pointer to the address data
 *          len  - length of the address data
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
xmlNewNodeNetAddr(char *tag, u_char *addr, int len)
{
    char *stag = "";
    int  i;

    static char netaddr_str[XML_TEMP_BUFFER_LEN];
    static char str[XML_TEMP_BUFFER_LEN];
    netaddr_str[0] = '\0';

    for ( i = 0; i < len; i++ )
    {
        snprintf(str, XML_TEMP_BUFFER_LEN, "%d", addr[i]);
        strcat(netaddr_str, stag); // separator
        strcat(netaddr_str, str);
        stag = ".";
    }

   return xmlNewNodeString( tag, netaddr_str );
}

/*----------------------------------------------------------------------------------------
 * Purpose: add a network address child node to a xml node
 * input:   parent_node - parent xml node
 *          addr - pointer to the address data
 *          len  - length of the address data
 * Output:  the child node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
xmlNewChildNetAddr(xmlNodePtr parent_node, char *tag, u_char *addr, int len)
{
    return xmlAddChild(parent_node, xmlNewNodeNetAddr(tag, addr, len));
}


/*----------------------------------------------------------------------------------------
 * Purpose: generate a xml node with SNPA value
 * input:   tag  - tag string
 *          addr - pointer to the address data
 *          len  - length of the address data
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
xmlNewNodeSNPA(char *tag, u_char *addr, int len)
{
    char *stag = "";
    int  i;
    static char snpa_str[XML_TEMP_BUFFER_LEN];
    static char str[XML_TEMP_BUFFER_LEN];
    snpa_str[0] = '\0';

    for ( i = 0; i < len; i++ )
    {
        snprintf( str, XML_TEMP_BUFFER_LEN, "%.2x", addr[i]);
        strcat(snpa_str, stag);
        strcat(snpa_str, str);
        stag = ":";
    }

    return xmlNewNodeString( tag, snpa_str );
}

/*----------------------------------------------------------------------------------------
 * Purpose: add a SNPA address child node to a xml node
 * input:   parent_node - parent xml node
 *          addr - pointer to the address data
 *          len  - length of the address data
 * Output:  the child node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
xmlNewChildSNPA(xmlNodePtr parent_node, char *tag, u_char *addr, int len)
{
    return xmlAddChild(parent_node, xmlNewNodeSNPA(tag, addr, len));
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate a xml node with AFI value and text
 * input:   tag  - tag string
 *          afi  - afi value
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 *
 * Ex: <AFI value="1">IPV4</AFI>
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
xmlNewNodeAFI(char *tag, int afi)
{
    xmlNodePtr node = NULL;
    char *afi_str   = "";

    switch(afi)
    {
        case BGP_AFI_IPv4:  afi_str = "IPV4"; break;
        case BGP_AFI_IPv6:  afi_str = "IPV6"; break;
        default: afi_str = "OTHER"; break;
    }

    node = xmlNewNodeString(tag, afi_str); 
    xmlNewPropInt(node, "value", afi);

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate a xml node with SAFI value and text
 * input:   tag  - tag string
 *          safi - safi value
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 *
 * Ex: <SAFI value="1">UNICAST</AFI>
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
xmlNewNodeSAFI(char *tag, int safi)
{
    xmlNodePtr node = NULL;
    char *safi_str   = "";

    switch(safi)
    {
        case BGP_MP_SAFI_UNICAST:    safi_str = "UNICAST"; break;
        case BGP_MP_SAFI_MULTICAST:  safi_str = "MULTICAST"; break;
        case BGP_MP_SAFI_MPLS:       safi_str = "MLPS"; break;
        default: safi_str = "OTHER"; break;
    }

    node = xmlNewNodeString(tag, safi_str); 
    xmlNewPropInt(node, "value", safi);

    return node;
}


/*----------------------------------------------------------------------------------------
 * Purpose: add PREFIX child nodes to a xml node
 * input:   parent_node - parent xml node
 *          prefix - pointer to the prefix
 *          len    - length of prefix
 *          afi    - address family of the prefix
 *          safi   - sub address family of the prefix
 *          lt     - pointer to an array of labels. it could be NULL.
 * Output:  the child node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
xmlNewChildPrefixes(xmlNodePtr parent_node, u_char *prefix, int len, u_int16_t afi, u_int8_t safi, u_char **lt )
{
    xmlNodePtr prefix_node = NULL;

    char *stag = "";
    int i, j, l;
    int bits;
    static char prefix_str[XML_TEMP_BUFFER_LEN];
    static char str[XML_TEMP_BUFFER_LEN];

    for ( i = 0; i < len; i = i + 1 + l )
    {
        /* Get prefix string */
        prefix_str[0] = '\0';
        bits = prefix[i];
        l = (bits + 7)/8;
        if ( l > 0 )
        {
            stag = ""; // no dot before first item in prefix
            for ( j = 0; j < l; j++ )
            {
                snprintf( str, XML_TEMP_BUFFER_LEN, "%d", prefix[i+1+j]);
                strcat(prefix_str, stag);
                strcat(prefix_str, str);
                stag = ".";
            }
        }
        else
            strcat(prefix_str, "0");

        snprintf( str, XML_TEMP_BUFFER_LEN, "/%d", bits );
        strcat(prefix_str, str);

        /* generate prefix node */
        prefix_node = xmlNewNodeString("PREFIX", prefix_str);

        /* add label */
        if (*lt != NULL) // add label
        {
            int label = **lt;
            *lt = *lt + 1;
            char *label_str;

            switch (label)
            {
                case BGPMON_LABEL_NULL:                label_str = "NULL";    break;
                case BGPMON_LABEL_WITHDRAW:            label_str = "WITH";    break;
                case BGPMON_LABEL_WITHDRAW_DUPLICATE:  label_str = "DUPW";    break;
                case BGPMON_LABEL_ANNOUNCE_NEW:        label_str = "NANN";    break;
                case BGPMON_LABEL_ANNOUNCE_DUPLICATE:  label_str = "DANN";    break;
                case BGPMON_LABEL_ANNOUNCE_DPATH:      label_str = "DPATH";   break;
                case BGPMON_LABEL_ANNOUNCE_SPATH:      label_str = "SPATH";   break;
                default: label_str = "UNKNOWN"; break;
            }
            xmlNewProp(prefix_node, BAD_CAST "label", BAD_CAST label_str);
        }
        xmlNewPropAFI(prefix_node,  afi);
        xmlNewPropSAFI(prefix_node, safi);
        xmlAddChild(parent_node, prefix_node);
    }

    return prefix_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate a xml node for state report 
 * input:   tag       - tag string
 *          stat_data - stat_data structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
xmlNewNodeStat(char *tag, stat_data_t *stat_data)
{
    xmlNodePtr node = NULL;

    /* Current value */
    node = xmlNewNodeInt(tag, stat_data->current);

    /* Currently only report the max and limit value */
    xmlNewPropInt(node, "max", stat_data->max);
    xmlNewPropInt(node, "limit", stat_data->limit);

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: add a child a xml node for state report 
 * input:   tag       - tag string
 *          stat_data - stat_data structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
inline xmlNodePtr
xmlNewChildStat(xmlNodePtr parent_node,  char *tag, stat_data_t *stat_data)
{
    return xmlAddChild(parent_node, xmlNewNodeStat(tag, stat_data));
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate a xml node for time report 
 * input:   tag       - tag string
 *          time_data - time_data structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
xmlNewNodeTime(char *tag, time_data_t *time_data)
{
    xmlNodePtr node = NULL;
    node = xmlNewNode(NULL, BAD_CAST tag);

    /* Current value */
    node = xmlNewNodeInt(tag, time_data->current);

    /* Currently only report the last_down and last_action time */
    xmlNewPropInt(node, "last_down",   time_data->last_down);
    xmlNewPropInt(node, "last_action", time_data->last_action);

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: add a child xml node for time report 
 * input:   tag       - tag string
 *          time_data - time_data structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
inline xmlNodePtr
xmlNewChildTime(xmlNodePtr parent_node,  char *tag, time_data_t *time_data)
{
    return xmlAddChild(parent_node, xmlNewNodeTime(tag, time_data));
}


/*----------------------------------------------------------------------------------------
 * Purpose: generate the OPT_PARAMETER node with child AUTHENTICATIOn/CAPABILITY node
 * input:    parms - parameters           
             len   - number of parameters 
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpOpenOptionalParameterNode(u_char *parms, int len )
{

    xmlNodePtr node       = NULL;

    if ( len > 0 )
    {
        /* OPT_PAR Node */
        node = xmlNewNode(NULL, BAD_CAST "OPT_PARAMETERS");

        /* Child nodes */
        int i;
        int l = 0;
        int type;
        for ( i = 0; i<len ; i = i + 2 + l )
        {
            xmlNodePtr par_node       = xmlNewNode(NULL, BAD_CAST "PARAMETER");
            xmlNodePtr par_type_node  = NULL;
            xmlNodePtr cap_node       = NULL;
            xmlNodePtr cap_type_node  = NULL;
            xmlNodePtr auth_node      = NULL;

            type = parms[i];
            l    = parms[i+1];


            switch ( type )
            {
                case BGP_OPEN_OPT_AUTH: // 1: authentication
                {
                    char *atag = "AUTHENTICATION";
                    par_type_node = xmlAddChild(par_node, xmlNewNodeString("TYPE", atag));
                    auth_node = xmlAddChild(par_node, xmlNewNode(NULL, BAD_CAST "AUTHENTICATION"));
                    break;
                }
                case BGP_OPEN_OPT_CAP:  // 2: capabilities
                {
                    char *atag = "CAPABILITIES";
                    par_type_node = xmlAddChild(par_node, xmlNewNodeString("TYPE", atag));
                    cap_node  = xmlAddChild(par_node, xmlNewNode(NULL, BAD_CAST "CAP"));

                    int capType = parms[i+1+1];
                    int capl    = parms[i+1+1+1];
                    switch ( capType ) /* Defined in bgppacket.h */
                    {
                        case multiprotocolCapability:
                        {
                            /* MULTIPROTOCOL */
                            char *atag = "MULTIPROTOCOL";
                            cap_type_node = xmlAddChild(cap_node, xmlNewNodeString("TYPE", atag));

                            xmlNodePtr mp_node = xmlNewChildString(cap_node, "MULTIPROTOCOL", "");
                            u_int16_t afi = ntohs( *((u_int16_t *) (parms+i+2+2) ) );
                            u_int8_t safi = parms[i+5+2];
                            xmlAddChild(mp_node, xmlNewNodeAFI("AFI",   afi));
                            xmlAddChild(mp_node, xmlNewNodeSAFI("SAFI", safi));
                            break;
                        }
                        case routeRefreshCapability:
                        {
                            /* ROUTE_REFRESH */
                            char *atag = "ROUTE_REFRESH";
                            cap_type_node = xmlAddChild(cap_node, xmlNewNodeString("TYPE", atag));
                            xmlNewChildString(cap_node, "ROUTE_REFRESH", "");
                            break;
                        }
                        case ciscoRefreshCapability:
                        {
                            /* CISCO_ROUTEREFRESH */
                            char *atag = "CISCO_ROUTE_REFRESH";
                            cap_type_node = xmlAddChild(cap_node, xmlNewNodeString("TYPE", atag));
                            xmlNewChildString(cap_node, "CISCO_ROUTE_REFRESH", "");
                            break;
                        }
                        // [EXTEND] additional capabilities
                        default:
                        {
                            char *atag = "OTHER";
                            cap_type_node = xmlAddChild(cap_node, xmlNewNodeString("TYPE", atag));
                            xmlNodePtr unknown_node = xmlNewChildString(cap_node, "OTHER", "");
                            if ( capl > 0 )
                            {
                                xmlNewChildOctets(unknown_node, "OCTETS", (parms+i+2+2), capl);
                            }
                            break;
                        }
                    }
                    xmlNewPropInt(cap_type_node, "value", capType);
                    break;
                }
                default:
                {
                    char *atag = "UNKNOWN";
                    par_type_node = xmlAddChild(par_node, xmlNewNodeString("TYPE", atag));
                    xmlNodePtr unknown_node = xmlNewChildString(par_node, "UNKNOWN", "");
                    if ( l > 0 )
                    {
                        xmlNewChildOctets(unknown_node, "OCTETS", (parms+i+1+1), l);
                    }
                    log_err("Unknown open parameter type!");
                    break;
                }
            }
            xmlNewPropInt(par_type_node, "value", type);

        }
    }

    return node;
}


/*----------------------------------------------------------------------------------------
 * Purpose: generate the WITHDRAWN node with child PREFIX nodes
 * input:   prefix  - pointer to the first prefix node
 *          afi     - afi value
 *          safi    - safi value
 *          lt      - pointer to an array of labels. it could be NULL
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 *
 * Ex: <WITHDRAWN count="2">
 *        <PREFIX label="WITH" afi="IPV4" afi_value="1" safi="UNICAST" safi_value="1">41.204.2/24</PREFIX>
 *        <PREFIX label="WITH" afi="IPV4" afi_value="1" safi="UNICAST" safi_value="1">63.82.84/24</PREFIX>
 *     </WITHDRAWN>
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genUpdateWithdrawnNode(u_char *prefix, int len, u_int16_t afi, u_int8_t safi, u_char **lt )
{
    xmlNodePtr node = xmlNewNode(NULL, BAD_CAST "WITHDRAWN");
    xmlNewChildPrefixes(node, prefix, len, afi, safi, lt);
    xmlNodePtr child_node = NULL;
    int count = 0;
    for (child_node=node->children; child_node!= NULL; child_node=child_node->next)
    {
        if (child_node->name && (strcmp((char*)child_node->name,"PREFIX") == 0))
        {
            count++;
        }
    }
    /* update the count of prefixes */
    xmlNewPropInt(node, "count", count);

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the NLRI node with child PREFIX nodes
 * input:   prefix  - pointer to the first prefix node
 *          afi     - afi value
 *          safi    - safi value
 *          lt      - pointer to an array of labels. it could be NULL
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 *
 * Ex: <NLRI count="2">
 *        <PREFIX label="NANN" afi="IPV4" afi_value="1" safi="UNICAST" safi_value="1">41.204.2/24</PREFIX>
 *        <PREFIX label="NANN" afi="IPV4" afi_value="1" safi="UNICAST" safi_value="1">63.82.84/24</PREFIX>
 *     </NLRI>
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genUpdateNlriNode(u_char *prefix, int len, u_int16_t afi, u_int8_t safi, u_char **lt )
{
    xmlNodePtr node = xmlNewNode(NULL, BAD_CAST "NLRI");
    xmlNewChildPrefixes(node, prefix, len, afi, safi, lt);
    xmlNodePtr child_node = NULL;
    int count = 0;
    for (child_node=node->children; child_node!= NULL; child_node=child_node->next)
    {
        if (child_node->name && (strcmp((char*)child_node->name,"PREFIX") == 0))
        {
            count++;
        }
    }
    /* update the count of prefixes */
    xmlNewPropInt(node, "count", count);
    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the flag node for an attribute
 * input:   flags   - flags in an interger
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 *
 * Ex:  <FLAGS>
 *         <TRANSITIVE/>
 *      </FLAGS>
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genAttributeFlagNode(int flags)
{
    xmlNodePtr node = xmlNewNode(NULL, BAD_CAST "FLAGS");
    //xmlNewPropInt(node, "code", flags);
    if ( (flags & BGP_ATTR_FLAG_OPTIONAL) > 0 ) { xmlAddChild(node, xmlNewNode(NULL, BAD_CAST "OPTIONAL"));   }
    if ( (flags & BGP_ATTR_FLAG_TRANS)    > 0 ) { xmlAddChild(node, xmlNewNode(NULL, BAD_CAST "TRANSITIVE")); }
    if ( (flags & BGP_ATTR_FLAG_PARTIAL)  > 0 ) { xmlAddChild(node, xmlNewNode(NULL, BAD_CAST "PARTIAL"));    }

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the ORIGIN attribute node
 * input:   origin  - origin value in an interger
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 *
 * Ex:  <ORIGIN value="0">IGP</ORIGIN>
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpOriginNode(BMF bmf, int origin)
{
    char *otag;
    int strlen = XML_TEMP_BUFFER_LEN;
    char str[strlen];

    snprintf( str, strlen, "%d", origin );
    switch ( origin )
    {
        case BGP_ORIGIN_IGP:
            otag = "IGP";
            break;
        case BGP_ORIGIN_EGP:
            otag = "EGP";
            break;
        case BGP_ORIGIN_INCOMPLETE:
            otag = "INCOMPLETE";
            break;
        default:
            otag = "OTHER";
            break;
    }
    xmlNodePtr node = xmlNewNodeString("ORIGIN", otag);
    xmlNewPropInt(node, "value", origin);

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the AS_SEG node
 * input:   value  - pointer to AS Segment data
 *          as_len - length of single AS number in bytes
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpASSegNode(u_char *value, int as_len)
{
    xmlNodePtr node = xmlNewNode(NULL, BAD_CAST "AS_SEG");

    static char str[XML_TEMP_BUFFER_LEN]; /* buffer for one single AS number */

    /* For each AS Segment */
    char *tag = NULL;
    int type = value[0];  /* Segment type */
    int l    = value[1];  /* Segment length */
    int i;
    str[0] = '\0';

    /* type */
    if      ( type == 1 ) tag = "AS_SET";
    else if ( type == 2 ) tag = "AS_SEQUENCE";
    else if ( type == 3 ) tag = "AS_CONFED_SEQUENCE";
    else if ( type == 4 ) tag = "AS_CONFED_SET";
    else                  tag = "OTHER";
    
    xmlNewPropString(node,  "type",   tag);
    xmlNewPropInt(node,     "length", l);

    /* AS */
    for ( i = 0 ; i < l ; i = i + 1 )
    {
        int as = 0;
		u_char *as_value = value + 2 + i*as_len;
        if      (as_len == 2) as += ntohs(*((u_int16_t *) (as_value))); /* 2-byte AS */
        else if (as_len == 4) as += ntohl(*((u_int32_t *) (as_value))); /* 4-byte AS */
        else                  as += ntohs(*((u_int16_t *) (as_value))); /* 2-byte AS */

        if(as >> 16) 
        {
            //sprintf(str, "%d.%d", (as >> 16) & 0xFFFF, as & 0xFFFF); /* DOT presentation */
            sprintf(str, "%d", as);                                    /* PLAIN presentation */
        } else {
            sprintf(str, "%d", as);
        }

        xmlNewChildString(node, "AS", str);
    }

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the AS_PATH attribute node
 * input:   value - pointer to AS path data
 *          len - length of AS path data
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 *
 * Ex:
 *      <AS_PATH>
 *        <AS_SEG type="AS_SEQUENCE" length="4">
 *          <AS>286</AS>
 *          <AS>3549</AS>
 *          <AS>6453</AS>
 *          <AS>9583</AS>
 *        </AS_SEG>
 *        <AS_SEG type="AS_SET" length="1">
 *          <AS>10380</AS>
 *        </AS_SEG>
 *      </AS_PATH>
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpASPathNode(u_char *value, int len, int asn_len)
{
    xmlNodePtr node = xmlNewNode(NULL, BAD_CAST "AS_PATH");

    int index = 0;
    while (index < len)
    {
        /* For each AS Segment */
        int l    = value[index + 1];  /* Segment length */

        xmlAddChild(node, genBgpASSegNode(value + index, asn_len));

        /* move the index to the start of next AS Segment */
        index = index + 2 + l*asn_len;
    }

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the AS_PATH attribute node
 * input:   value - pointer to AS path data
 *          len - length of AS path data
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpAS4PathNode(u_char *value, int len )
{
    xmlNodePtr node = xmlNewNode(NULL, BAD_CAST "AS4_PATH");

    int index = 0;
    while (index < len)
    {
        /* For each AS Segment */
        int l    = value[index + 1];  /* Segment length */

        xmlAddChild(node, genBgpASSegNode(value + index, 4)); /* 4-byte AS */

        /* move the index to the start of next AS Segment */
        index = index + 2 + l*4;
    }

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the COMMUNITIES attribute node
 * input:   list - pointer to a list of BGP communities
 *          len - length of BGP communities
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 *
 * Ex:
 *  <COMMUNITIES>
 *     <COMMUNITY>
 *        <AS>286</AS>
 *        <VALUE>19</VALUE>
 *     </COMMUNITY>
 *     <COMMUNITY>
 *       <AS>286</AS>
 *       <VALUE>29</VALUE>
 *     </COMMUNITY>
 *  </COMMUNITIES>
 *
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpCommunitiesNode(BMF bmf, u_char *list, int len )
{
    xmlNodePtr node = xmlNewNode(NULL, BAD_CAST "COMMUNITIES");

    char *ctag = "COMMUNITY";
    char *rtag = "RESERVED_COMMUNITY";
    char *atag = "AS";
    char *vtag = "VALUE";
    int i;

    for ( i = 0; i < len; i = i + 4 )
    {
        u_int16_t as  = ntohs( *((u_int16_t *) list));
        u_int16_t val = ntohs(*((u_int16_t *) (list+2)));

        if ( as == 0xFFFF && val ==  0xFF01 )
        {
            xmlAddChild(node, xmlNewNode(NULL, BAD_CAST "NO_EXPORT"));
        }
        else if ( as == 0xFFFF && val == 0xFF02 )
        {
            xmlAddChild(node, xmlNewNode(NULL, BAD_CAST "NO_ADVERTISE"));
        }
        else if ( as == 0xFFFF && val == 0xFF03 )
        {
            xmlAddChild(node, xmlNewNode(NULL, BAD_CAST "NO_EXPORT_SUBCONFED"));
        }
        else if ( as == 0x0000 || as == 0xFFFF )
        {
            xmlNodePtr child_node = xmlNewNode(NULL, BAD_CAST rtag);
            xmlNewChildInt(child_node, atag, as);
            xmlNewChildInt(child_node, vtag, val);
            xmlAddChild(node, child_node);
        }
        else
        {
            xmlNodePtr child_node = xmlNewNode(NULL, BAD_CAST ctag);
            xmlNewChildInt(child_node, atag, as);
            xmlNewChildInt(child_node, vtag, val);
            xmlAddChild(node, child_node);
        }
        list = list+4;
    }

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the CLUSTER_LIST attribute node
 * input:   value - pointer to cluster list data
 *          len - length of cluster list data
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpClusterListNode(u_char *value, int len )
{
    xmlNodePtr node = xmlNewNode(NULL, BAD_CAST "CLUSTER_LIST");

    int index = 0;
    while (index < len)
    {
        /* Each ID is a 4-byte IP */
        xmlNewChildIP(node, "ID", *((u_int32_t *) value + index ));

        /* Move the index to the next cluster id */
        index = index + 32; /* advance the index by 4-byte */
    }

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the MP_REACH_NLRI attribute node
 * input:   attr - pointer to mpreach attribute
 *          len  - length of mpreach attribute
 *          lt   - pointer to an array of labels. it could be NULL
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 *
 * Ex:
 *  <MP_REACH_NLRI>
 *     <AFI value="1">IPV4</AFI>
 *     <SAFI value="2">MULTICAST</SAFI>
 *     <NEXT_HOP>129.250.0.11</NEXT_HOP>
 *     <SNPA_LIST_LEN>0</SNPA_LIST_LEN>
 *     <SNPA_LIST count="0"/>
 *     <NLRI count="1">
 *        <PREFIX label="NULL" afi="IPV4" afi_value="1" safi="MULTICAST" safi_value="2">208.28.235/24</PREFIX>
 *     </NLRI>
 *  </MP_REACH_NLRI>
 *
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpMPReachNode(u_char *attr, int len, u_char **lt)
{
    xmlNodePtr node = xmlNewNode(NULL, BAD_CAST "MP_REACH_NLRI");

    u_char *snpa;
    int i;
    int m;
    u_int16_t afi =  ntohs(*((u_int16_t *) attr));
    u_int8_t safi =  *((u_int8_t *) (attr+2));

    /* afi / safi */
    xmlAddChild(node, xmlNewNodeAFI("AFI", afi));
    xmlAddChild(node, xmlNewNodeSAFI("SAFI", safi));

    /* next hop */
    int nhlen = *(u_int8_t *) (attr+3);
    xmlAddChild(node, xmlNewNodeNetAddr("NEXT_HOP", attr+4, nhlen));

    /* length */
    m =  *(u_int8_t *) (attr + 4 + nhlen );
    xmlNewChildInt(node, "SNPA_LIST_LEN", m);


    xmlNodePtr snpa_list_node = xmlAddChild(node, xmlNewNode(NULL, BAD_CAST "SNPA_LIST"));
    snpa = (attr + 5 + nhlen );
    for ( i = 0; i < m; i ++ )
    {
        xmlAddChild(snpa_list_node, xmlNewNodeSNPA("SNPA", snpa+1, *snpa));
        snpa = snpa + 1 + *snpa;
    }
    xmlNewPropInt(snpa_list_node, "count", m);

    /* nlri */
    xmlAddChild(node, genUpdateNlriNode(snpa, len - (snpa-attr),  afi, safi, lt));

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the MP_UNREACH_NLRI attribute node
 * input:   attr - pointer to mpreach attribute
 *          len  - length of mpreach attribute
 *          lt   - pointer to an array of labels. it could be NULL
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 *
 * Ex:
 *
 *  <MP_UNREACH_NLRI>
 *     <AFI value="1">IPV4</AFI>
 *     <SAFI value="2">MULTICAST</SAFI>
 *     <WITHDRAWN count="1">
 *       <PREFIX label="NULL" afi="IPV4" afi_value="1" safi="MULTICAST" safi_value="2">208.28.235/24</PREFIX>
 *     </WITHDRAWN>
 *  </MP_UNREACH_NLRI>
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpMPUnreachNode(u_char *attr, int len, u_char **lt)
{
    xmlNodePtr node = xmlNewNode(NULL, BAD_CAST "MP_UNREACH_NLRI");

    u_int16_t afi =  ntohs(*((u_int16_t *) attr));
    u_int8_t safi =  *((u_int8_t *) (attr+2));

    /* afi / safi */
    xmlAddChild(node, xmlNewNodeAFI("AFI", afi));
    xmlAddChild(node, xmlNewNodeSAFI("SAFI", safi));

    /* withdrawn */
    xmlAddChild(node, genUpdateWithdrawnNode(attr+3, len-3, afi, safi, lt));

    return node;
}


/*----------------------------------------------------------------------------------------
 * Purpose: generate the PATH_ATTRIBUTES node
 * input:   attr - pointer to normal attribute
 *          len  - length of normal attribute
 *          lt   - pointer to an array of labels. it could be NULL
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genUpdateAttributesNode(BMF bmf, u_char *attr, int len, u_char **lt )
{
    xmlNodePtr attributes_node = NULL;
    attributes_node = xmlNewNode(NULL, BAD_CAST "PATH_ATTRIBUTES");

    int i;
    int hl;
    int l;
    int flags;
    int type;
    int count = 0;
    u_char *value;

    /* Session specific setting */
//    Session_structp sp = getSessionByID(bmf->sessionID);
//    int asn_len = sp->fsm.ASNumlen;

    /* For each attribute */
    for ( i = 0; i < len; i = i + 2 + hl + l )
    {
        xmlNodePtr attr_node = xmlNewNode(NULL, BAD_CAST "ATTRIBUTE");

        /* flags */
        flags = attr[i];


        type  = attr[i+1];


		
        if ( (flags & 0x10) > 0 )
        {

            hl = 2;
            l = ntohs( *((u_int16_t *) (attr+i+2)) );
            value = attr+i+4;
        }
        else
        {
            hl = 1;
            l = attr[i+2];

            value = attr+i+3;
        }
        xmlAddChild(attr_node, genAttributeFlagNode(flags));

        /* process each attribute */
        xmlNodePtr length_node = NULL;
        xmlNodePtr type_node   = NULL;
        length_node = xmlAddChild(attr_node, xmlNewNodeInt("LENGTH", l));
        switch ( type ) /* Defined in bgpmessagetypes.h */
        {
            case BGP_ATTR_ORIGIN:
            {
                /* ORIGIN(1) */
                char *atag = "ORIGIN";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlAddChild(attr_node, genBgpOriginNode(bmf, value[0]));
                break;
            }
            case BGP_ATTR_AS_PATH:
            {
                /* AS PATH(2) */
                char *atag = "AS_PATH";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
//				printf("\n Value is %02x, len is %02x, asn_len is %d", value, l, asn_len);

                xmlAddChild(attr_node, genBgpASPathNode(value, l, /*asn_len*/ 2));
                break;
            }
            case BGP_ATTR_NEXT_HOP:
            {
                /* NEXT HOP(3) */
                char *atag = "NEXT_HOP";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlAddChild(attr_node, xmlNewNodeIP(atag, *((u_int32_t *) value )));
                break;
            }
            case BGP_ATTR_MULTI_EXIT_DISC:
            {
                /* MED(4) */
                char *atag = "MULTI_EXIT_DISC";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlAddChild(attr_node, xmlNewNodeInt(atag, ntohl( *((u_int32_t *) value ))));
                break;
            }
            case BGP_ATTR_LOCAL_PREF:
            {
                /* LOCAL_PREF(5) */
                char *atag = "LOCAL_PREF";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlAddChild(attr_node, xmlNewNodeInt(atag, ntohl(*((u_int32_t *) value ))));
                break;
            }
            case BGP_ATTR_ATOMIC_AGGREGATE:
            {
                /* ATOMIC_AGGREGATE(6) */
                char *atag = "ATOMIC_AGGREGATE";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlAddChild(attr_node, xmlNewNode(NULL, BAD_CAST atag));
                break;
            }
            case BGP_ATTR_AGGREGATOR:
            {
                /* AGGREGATOR(7) */
                char *atag = "AGGREGATOR";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlNodePtr temp_node = xmlAddChild(attr_node, xmlNewNode(NULL, BAD_CAST atag));
                xmlNewChildInt(temp_node, "AS",   ntohs( *((u_int16_t *) value ))); /* 2-byte AS */
                xmlNewChildIP(temp_node,  "ADDR", *((u_int32_t *) value+2 ));       /* 4-byte address */
                break;
            }
            case BGP_ATTR_COMMUNITIES: // RFC 1997
            {
                /* COMMUNITIES(8) */
                char *atag = "COMMUNITIES";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlAddChild(attr_node, genBgpCommunitiesNode(bmf, value, l));
                break;
            }
            case BGP_ATTR_ORIGINATOR_ID: // RFC 2796
            {
                /* ORIGINATOR_ID(9) */
                char *atag = "ORIGINATOR_ID";
                type_node = xmlAddChild(attr_node,  xmlNewNodeString("TYPE", atag));
                xmlNewChildIP(attr_node, atag, *((u_int32_t *) value ));
                break;
            }
            case BGP_ATTR_CLUSTER_LIST: // RFC 2796
            {
                /* CLUSTER_LIST(10) */
                char *atag = "CLUSTER_LIST";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlAddChild(attr_node, genBgpClusterListNode(value, l));
                break;
            }
            case BGP_ATTR_DPA: // expired!!
            {
                /* DESTINATION_PREFERENCE */
                char *atag = "DESTINATION_PREFERENCE";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlNodePtr temp_node = xmlAddChild(attr_node, xmlNewNode(NULL, BAD_CAST atag));
                if ( l > 0 ) {
                    xmlNewChildOctets(temp_node, "OCTETS", value, l);
                }
                break;
            }
            case BGP_ATTR_ADVERTISER: // Historic!!
            {
                /* ADVERTISER(12) */
                char *atag = "ADVERTISER";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlNodePtr temp_node = xmlAddChild(attr_node, xmlNewNode(NULL, BAD_CAST atag));
                if ( l > 0 ) {
                    xmlNewChildOctets(temp_node, "OCTETS", value, l);
                }
                break;
            }
            case BGP_ATTR_RCID_PATH: //  Historic!!
            {
                /* RCID_PATH(13) */
                char *atag = "RCID_PATH";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlNodePtr temp_node = xmlAddChild(attr_node, xmlNewNode(NULL, BAD_CAST atag));
                if ( l > 0 ) {
                    xmlNewChildOctets(temp_node, "OCTETS", value, l);
                }
                break;
            }
            case BGP_ATTR_MP_REACH_NLRI: // RFC 2858
            {
                /* MP_REACH_NLRI(14) */
                char *atag = "MP_REACH_NLRI";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlAddChild(attr_node, genBgpMPReachNode(value, l, lt));
                break;
            }
            case BGP_ATTR_MP_UNREACH_NLRI: // RFC 2858
            {
                /* MP_UNREACH_NLRI(15) */
                char *atag = "MP_UNREACH_NLRI";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlAddChild(attr_node, genBgpMPUnreachNode(value, l, lt));
                break;
            }
            case BGP_ATTR_EXT_COMMUNITIES: // RFC 4360
            {
                /* EXTENDED_COMMUNITIES(16) */
                // The semantic of "Extended Communities" is determined by the value of "type" and "subtype" 
                // Currently we only preserve the OCTETS string
                char *atag = "EXTENDED_COMMUNITIES";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlNodePtr temp_node = xmlAddChild(attr_node, xmlNewNode(NULL, BAD_CAST atag));
                if ( l > 0 ) {
                    xmlNewChildOctets(temp_node, "OCTETS", value, l);
                }
                break;
            }
            case BGP_ATTR_AS4_PATH: // RFC 4893
            {
                /* AS4_PATH(17) */
                char *atag = "AS4_PATH";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlAddChild(attr_node, genBgpAS4PathNode(value, l));
                break;
            }
            case BGP_ATTR_AS4_AGGREGATOR: // RFC 4893
            {
                /* AS4_AGGREGATOR(18) */
                char *atag = "AS4_AGGREGATOR";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlNodePtr temp_node = xmlAddChild(attr_node, xmlNewNode(NULL, BAD_CAST atag));
                xmlNewChildInt(temp_node, "AS",   ntohl( *((u_int32_t *) value ))); /* 4-byte AS */
                xmlNewChildIP(temp_node,  "ADDR", *((u_int32_t *) value+4 ));       /* 4-byte address */
                break;
            }
            //[EXTEND] Additional attributes here
            default:
            {
                char *atag = "OTHER";
                type_node = xmlAddChild(attr_node, xmlNewNodeString("TYPE", atag));
                xmlNodePtr other_node = xmlAddChild(attr_node, xmlNewNode(NULL, BAD_CAST atag));
                if ( l > 0 ) {
                    xmlNewChildOctets(other_node, "OCTETS", value, l);
                }
                break;
            }
        }
        xmlNewPropInt(type_node, "value", type);
        xmlAddChild(attributes_node, attr_node);
        count++;
    }
    xmlNewPropInt(attributes_node, "count", count);

    return attributes_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate OPEN node
 * input:   bmf - pointer to BMF structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpOpenNode(BMF bmf)
{
    xmlNodePtr bgp_node = NULL;
    PBgpOpen open = (PBgpOpen)(bmf->message + BGP_HEADER_LEN);

    /* BGP OPEN node */
    bgp_node = xmlNewNode(NULL, BAD_CAST "OPEN");

    /* Child nodes */
    /* VERSION     */ xmlNewChildInt(bgp_node, "VERSION",     getBGPOpenVersion(open));
    /* SRC_AS      */ xmlNewChildInt(bgp_node, "SRC_AS",      getBGPOpenAutonomousSystem(open));
    /* HOLD_TIME   */ xmlNewChildInt(bgp_node, "HOLD_TIME",   getBGPOpenHoldTime(open));
    /* SRC_BGP     */ xmlNewChildIP(bgp_node,  "SRC_BGP",     getBGPOpenIdentifier(open));
    /* OPT_PAR_LEN */ xmlNewChildInt(bgp_node, "OPT_PAR_LEN", getBGPOpenOptionalParameterLength(open));
    /* OPT_PAR     */ xmlAddChild(bgp_node, genBgpOpenOptionalParameterNode(/* params */ open->optionalParameter,
                                                                            /* length */ getBGPOpenOptionalParameterLength(open)));
    return bgp_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate UPDATE node
 * input:   bmf - pointer to BMF structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpUpdateNode(BMF bmf)
{
    xmlNodePtr bgp_node = NULL;
    /* BGP Update node */
    bgp_node = xmlNewNode(NULL, BAD_CAST "UPDATE");

    /* header */
    PBgpHeader hdr = NULL;
    u_int16_t  bgpType   = 0;
    u_int32_t  bgpMsgLen = 0;

    hdr = (PBgpHeader)(bmf->message);
    bgpType   = hdr->type;
    bgpMsgLen = getBGPHeaderLength(hdr);

    u_char *lt1 = NULL;
    if ( bmf->type == BMF_TYPE_MSG_LABELED ) lt1 = (u_char *)(bmf->message + bgpMsgLen);

    u_char *update = bmf->message + BGP_HEADER_LEN;
    int len        = bgpMsgLen   - BGP_HEADER_LEN;
    int real_len   = bmf->length - BGP_HEADER_LEN; /* In case of prefixes have been truncated */
    u_char **lt    = &lt1;

    /* len */
    int wlen = ntohs( *((u_int16_t *) update)  );           if (wlen > real_len - 4) wlen = real_len - 4; /* prefixes are truncated */
    int alen = ntohs( *((u_int16_t *) (update+2+wlen)) );
    int nlen = len - alen - wlen - 4;                       if (nlen > real_len - 4) nlen = real_len - 4; /* prefixes are truncated */

    /* Child nodes */
    /* WITHDRAWN_LEN  */ xmlAddChild(bgp_node, xmlNewNodeInt("WITHDRAWN_LEN", wlen));
    /* WITHDRAWN      */ xmlAddChild(bgp_node, genUpdateWithdrawnNode(update+2, wlen, 1, 1, lt));
    /* ATTRIBUTES_LEN */ xmlAddChild(bgp_node, xmlNewNodeInt("PATH_ATTRIBUTES_LEN", alen));
    /* ATTRIBUTES     */ xmlAddChild(bgp_node, genUpdateAttributesNode(bmf, update+4+wlen, alen, lt));
    /* NLRI           */ xmlAddChild(bgp_node, genUpdateNlriNode(update+4+wlen+alen, nlen, 1, 1, lt));

    return bgp_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate TIME node
 * input:   bmf - pointer to BMF structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 *
 * Ex:
 *  <TIME>
 *    <TIMESTAMP>1229905411</TIMESTAMP>
 *    <DATETIME>2008-12-22T00:23:31Z</DATETIME>
 *    <PRECISION_TIME>300</PRECISION_TIME>
 *  </TIME>
 * -------------------------------------------------------------------------------------*/
xmlNodePtr genTimeNode(BMF bmf)
{
    xmlNodePtr time_node = NULL;

    /* TIME node */
    time_node = xmlNewNode(NULL, BAD_CAST "TIME");

    /* Child nodes */
    /* TIMESTAMP      */                             xmlNewChildUnsignedInt(time_node,  "TIMESTAMP",      bmf->timestamp);
    /* DATETIME       */ if (GMT_TIME_STAMP == TRUE) xmlNewChildGmtTime(time_node,      "DATETIME",       bmf->timestamp);
    /* PRECISION_TIME */                             xmlNewChildUnsignedInt(time_node,  "PRECISION_TIME", bmf->precisiontime);

    return time_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate PEERING node
 * input:   bmf - pointer to BMF structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 *
 * Ex:
 *  <PEERING>
 *    <SRC_ADDR afi="IPV4" afi_value="1" if_index="0">208.51.134.246</SRC_ADDR>
 *    <SRC_PORT>0</SRC_PORT>
 *    <SRC_AS>3549</SRC_AS>
 *    <DST_ADDR afi="IPV4" afi_value="1" if_index="0">128.223.51.102</DST_ADDR>
 *    <DST_PORT>0</DST_PORT>
 *    <DST_AS>6447</DST_AS>
 *    <BGPID>128.223.51.102</BGPID>
 *  </PEERING>
 * -------------------------------------------------------------------------------------*/
xmlNodePtr genPeeringNode(BMF bmf)
{
    xmlNodePtr peering_node = NULL;
    peering_node = xmlNewNode(NULL, BAD_CAST "PEERING");

    char *srcAddr, *dstAddr;
    u_int16_t srcPort, srcAS, dstPort, dstAS;

    xmlNodePtr src_addr_node, dst_addr_node;

    /* PEERING NODE */
    Session_structp sp = getSessionByID(bmf->sessionID);

    if (sp)
    {
        // Have to know how to directly access srcaddr, srcport, arcas, dstaddr, dstport, dstas
        srcAddr = sp->sessionRealSrcAddr;    // Use real source address
        srcPort = sp->configInUse.localPort;
        srcAS   = sp->configInUse.localAS2;
        dstAddr = sp->configInUse.remoteAddr;
        dstPort = sp->configInUse.remotePort;
        dstAS   = sp->configInUse.remoteAS2;

        /* SRC_ADDR */ src_addr_node = xmlNewChildString(peering_node, "SRC_ADDR", srcAddr);
        /* SRC_PORT */ xmlNewChildInt(peering_node,    "SRC_PORT", srcPort);
        /* SRC_AS   */ xmlNewChildInt(peering_node,    "SRC_AS",   srcAS);
        /* DST_ADDR */ dst_addr_node = xmlNewChildString(peering_node, "DST_ADDR", dstAddr);
        /* DST_PORT */ xmlNewChildInt(peering_node,    "DST_PORT", dstPort);
        /* DST_AS   */ xmlNewChildInt(peering_node,    "DST_AS",   dstAS);
        /* BGPID    */ xmlNewChildBGPID(peering_node,  "BGPID",    sp->configInUse.remoteBGPID);
	/* AS Num Len*/xmlNewChildInt(peering_node,    "AS_NUM_LEN",   sp->fsm.ASNumlen);

        xmlNewPropAFI(src_addr_node, get_afi(srcAddr));
        xmlNewPropInt(src_addr_node, "if_index", 0);
        xmlNewPropAFI(dst_addr_node, get_afi(dstAddr));
        xmlNewPropInt(dst_addr_node, "if_index", 0);
    }
    //printNode(peering_node);
    return peering_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate NOTIFICATION node
 * input:   bmf - pointer to BMF structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpNotificationNode(BMF bmf)
{
    xmlNodePtr node = NULL;
    PBgpNotification ntf = (PBgpNotification)(bmf->message + BGP_HEADER_LEN);

    char *code_str    = NULL;
    char *subcode_str = NULL;
    

    switch(ntf->errorCode)
    {
        case 1:
        {
            code_str = "Message Header Error"; 
            switch(ntf->errorSubcode)
            {
                case 1: { subcode_str = "Connection Not Synchronized"; break; }
                case 2: { subcode_str = "Bad Message Length";          break; }
                case 3: { subcode_str = "Bad Message Type";            break; }
            }
            break;
        }
        case 2:
        {
            code_str = "OPEN Message Error"; 
            switch(ntf->errorSubcode)
            {
                case 1: { subcode_str = "Unsupported Version Number";         break; }
                case 2: { subcode_str = "Bad Peer AS";                        break; }
                case 3: { subcode_str = "Bad BGP Identifier";                 break; }
                case 4: { subcode_str = "Unsupported Optional Parameter";     break; }
                case 5: { subcode_str = "Authentication Failure";             break; }
                case 6: { subcode_str = "Unacceptable Hold Time";             break; }
            }
            break;
        }
        case 3:
        {
            code_str = "UPDATE Message Error"; 
            switch(ntf->errorSubcode)
            {
                case 1:  { subcode_str = "Malformed Attribute List";          break; }
                case 2:  { subcode_str = "Unrecognized Well-known Attribute"; break; }
                case 3:  { subcode_str = "Missing Well-known Attribute";      break; }
                case 4:  { subcode_str = "Attribute Flags Error";             break; }
                case 5:  { subcode_str = "Attribute Length Error";            break; }
                case 6:  { subcode_str = "Invalid ORIGIN Attribute";          break; }
                case 7:  { subcode_str = "AS Routing Loop";                   break; }
                case 8:  { subcode_str = "Invalid NEXT_HOP Attribute";        break; }
                case 9:  { subcode_str = "Optional Attribute Error";          break; }
                case 10: { subcode_str = "Invalid Network Field";             break; }
                case 11: { subcode_str = "Malformed AS_PATH";                 break; }
            }
            break;
        }
        case 4:
        {
            code_str = "Hold Timer Expired"; 
            break;
        }
        case 5:
        {
            code_str = "Finite State Machine Error"; 
            break;
        }
        case 6:
        {
            code_str = "Cease";
            switch(ntf->errorSubcode)
            {
                case 1: { subcode_str = "Maximum Number of Prefixes Reached"; break; }
                case 2: { subcode_str = "Administrative Shutdown";            break; }
                case 3: { subcode_str = "Peer De-configured";                 break; }
                case 4: { subcode_str = "Administrative Reset";               break; }
                case 5: { subcode_str = "Connection Rejected";                break; }
                case 6: { subcode_str = "Other Configuration Change";         break; }
                case 7: { subcode_str = "Connection Collision Resolution";    break; }
            }
            break;
        }
        default:
        {
            break;
        }

    }

    /* NOTIFICATION node */
    node = xmlNewNode(NULL, BAD_CAST "NOTIFICATION");

    /* Child nodes */
    xmlNewPropInt(xmlNewChildString(node,    "CODE",    code_str   ), "value", ntf->errorCode);
    xmlNewPropInt(xmlNewChildString(node,    "SUBCODE", subcode_str), "value", ntf->errorSubcode);

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate KEEPALIVE node
 * input:   bmf - pointer to BMF structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpKeepaliveNode(BMF bmf)
{
    xmlNodePtr node = NULL;
    node = xmlNewNode(NULL, BAD_CAST "KEEPALIVE");
    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate ROUTE_REFRESH node
 * input:   bmf - pointer to BMF structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpRouteRefreshNode(BMF bmf)
{
    xmlNodePtr node = NULL;
    node = xmlNewNode(NULL, BAD_CAST "ROUTE_REFRESH");
    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate CISCO_ROUTE_REFRESH node
 * input:   bmf - pointer to BMF structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpCiscoRouteRefreshNode(BMF bmf)
{
    xmlNodePtr node = NULL;
    node = xmlNewNode(NULL, BAD_CAST "CISCO_ROUTE_REFRESH");
    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate STATE_CHANGE node
 * input:   bmf - pointer to BMF structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpStateChangeNode(BMF bmf)
{
    xmlNodePtr state_change_node = NULL;
    StateChangeMsg *sc    = NULL;

    sc = (StateChangeMsg *)(bmf->message);

    /* STATE_CHANGE node */ state_change_node = xmlNewNode(NULL, BAD_CAST "STATE_CHANGE");

    /* Child nodes */
    /* OLD_STATE */ xmlNewChildInt(state_change_node, "OLD_STATE",   sc->oldState);
    /* NEW_STATE */ xmlNewChildInt(state_change_node, "NEW_STATE",   sc->newState);
    /* REASON    */ xmlNewChildInt(state_change_node, "REASON",      sc->reason);

    return state_change_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate UNKNOWN node
 * input:   bmf - pointer to BMF structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpUnknownNode(BMF bmf)
{
    xmlNodePtr node = NULL;

    PBgpHeader hdr = NULL;
    u_int16_t  bgpType = 0;
    u_int32_t  bgpMsgLen = 0;

    hdr       = (PBgpHeader)(bmf->message);
    bgpType   = hdr->type;
    bgpMsgLen = getBGPHeaderLength(hdr);

    node = xmlNewNode(NULL, BAD_CAST "UNKNOWN");
    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate MARKER node
 * input:   hdr - PBgpHeader sturcture
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 *
 * Ex:
 *  <MARKER length="16">FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF</MARKER>
 * -------------------------------------------------------------------------------------*/

xmlNodePtr
genAsciiMsgMarkerNode(PBgpHeader hdr)
{
    xmlNodePtr node = NULL;

    if (hdr)
    {
        /* MARKER node */
        if (static_marker_node == NULL)
        {
            static_marker_node = xmlNewNodeOctets("MARKER", hdr->mask, sizeof(hdr->mask));
        }
        node = xmlCopyNode(static_marker_node, 1);
    }

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate LENGTH node
 * input:   hdr - PBgpHeader sturcture
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 *
 * Ex:
 *  <LENGTH>34</LENGTH>
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genAsciiMsgLengthNode(PBgpHeader hdr)
{
    xmlNodePtr node = NULL;

    if (hdr)
    {
        /* LENGTH node */
        int bgpMsgLen = getBGPHeaderLength(hdr);
        node = xmlNewNodeInt("LENGTH", bgpMsgLen);
    }

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate TYPE node
 * input:   hdr - PBgpHeader sturcture
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 *
 * <TYPE value="2">UPDATE</TYPE>
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genAsciiMsgTypeNode(PBgpHeader hdr)
{
    xmlNodePtr node = NULL;

    if (hdr)
    {
        u_int16_t bgpType = hdr->type;
        char *bgpTypeStr = NULL;

        /* Choice of message types */
        switch ( bgpType )
        {
            case typeOpen:              { /* OPEN                */ bgpTypeStr = "OPEN";                break; }
            case typeUpdate:            { /* UPDATE              */ bgpTypeStr = "UPDATE";              break; }
            case typeNotification:      { /* NOTIFICATION        */ bgpTypeStr = "NOTIFICATION";        break; }
            case typeKeepalive:         { /* KEEPALIVE           */ bgpTypeStr = "KEEPALIVE";           break; }
            case typeRouteRefresh:      { /* ROUTE_REFRESH       */ bgpTypeStr = "ROUTE_REFRESH";       break; }
            case typeRouteRefreshCisco: { /* CISCO_ROUTE_REFRESH */ bgpTypeStr = "CISCO_ROUTE_REFRESH"; break; }
            default:                    { /* UNKNOWN             */ bgpTypeStr = "UNKNOWN";             break;  }
        }

        /* TYPE node */
        node = xmlNewNodeString("TYPE", bgpTypeStr);
        xmlNewPropInt(node, "value", hdr->type);
    }

    return node;
}


/*----------------------------------------------------------------------------------------
 * Purpose: generate ASCII_MSG node
 * input:   bmf - our internal BMF message
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genAsciiMsgNode(BMF bmf)
{
    xmlNodePtr ascii_node = NULL;

    if (ASCII_MESSAGES == TRUE)
    {

        PBgpHeader hdr       = NULL;
        u_int16_t  bgpType   = 0;
        u_int32_t  bgpMsgLen = 0;

        hdr       = (PBgpHeader)(bmf->message);
        bgpType   = hdr->type;
        bgpMsgLen = getBGPHeaderLength(hdr);


        /* ASCII_MSG node */
        ascii_node = xmlNewNode(NULL, BAD_CAST "ASCII_MSG");

        xmlAddChild(ascii_node, genAsciiMsgMarkerNode(hdr));
        xmlAddChild(ascii_node, genAsciiMsgLengthNode(hdr));
        xmlAddChild(ascii_node, genAsciiMsgTypeNode(hdr));

        /* Choice of message types */
        switch ( bgpType )
        {
            case typeOpen:
            {
                /* OPEN */
                xmlAddChild(ascii_node, genBgpOpenNode(bmf));
                break;
            }
            case typeUpdate:
            {
                /* UPDATE */
                xmlAddChild(ascii_node, genBgpUpdateNode(bmf));
                break;
            }
            case typeNotification:
            {
                /* NOTIFICATION */
                xmlAddChild(ascii_node, genBgpNotificationNode(bmf));
                break;
            }
            case typeKeepalive:
            {
                /* KEEPALIVE */
                xmlAddChild(ascii_node, genBgpKeepaliveNode(bmf));
                break;
            }
            case typeRouteRefresh:
            {
                /* ROUTE_REFRESH */
                xmlAddChild(ascii_node, genBgpRouteRefreshNode(bmf));
                break;
            }
            case typeRouteRefreshCisco:
            {
                /* CISCO_ROUTE_REFRESH */
                xmlAddChild(ascii_node, genBgpCiscoRouteRefreshNode(bmf));
                break;
            }
            default:
            {
                /* UNKNOWN */
                xmlAddChild(ascii_node, genBgpUnknownNode(bmf));
                break;
            }
        }

        /* Attributes */
        //xmlNewPropInt(ascii_node, "type",   bgpType);
    }
    return ascii_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate OCTET_MSG node
 * input:   bmf - our internal BMF message
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genOctetMsgNode(BMF bmf)
{
    xmlNodePtr octet_node = NULL;
    PBgpHeader hdr        = NULL;
    u_int16_t  bgpType    = 0;
    u_int32_t  bgpMsgLen  = 0;

    // Due to the labels at the end of BMF, we needs to know the real length of raw bgp update
    hdr       = (PBgpHeader)(bmf->message);
    bgpType   = hdr->type;
    bgpMsgLen = getBGPHeaderLength(hdr);

    /* OCTET_MSG node */
    octet_node = xmlNewNode(NULL, BAD_CAST "OCTET_MSG");
    xmlAddChild(octet_node, genAsciiMsgMarkerNode(hdr));
    xmlAddChild(octet_node, genAsciiMsgLengthNode(hdr));
    xmlAddChild(octet_node, genAsciiMsgTypeNode(hdr));
    xmlNewChildOctets(octet_node, "OCTETS", bmf->message, bgpMsgLen);

    return octet_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: convert the queue status message to an XML text representation
 * input:	queueName - queue name
 * Output: the position of pointer in the buffer
 * He Yan @ Jun 22, 2008
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/ 
xmlNodePtr
genPacingNode(char *queueName)
{
    xmlNodePtr node = NULL;
    /* QUEUE node */ node = xmlNewNode(NULL, BAD_CAST "PACING");

    stat_data_t pc_data, wl_data;
    memset(&pc_data, 0, sizeof(stat_data_t));
    memset(&wl_data, 0, sizeof(stat_data_t));

    // Accumulated and current count
    pc_data.accu    = getPacingCount(queueName);
    pc_data.current = getCurrentPacingCount(queueName);
    wl_data.current = getCurrentWritesLimit(queueName);

    int   flag     = getPacingState(queueName);
    char* flag_str = NULL;
    if (flag == TRUE) { flag_str = "1"; }
    else              { flag_str = "0"; }

    xmlNewChildString(node, "FLAG",         flag_str);
    xmlNewChildStat(node,   "COUNT",        &pc_data);
    xmlNewChildStat(node,   "WRITE_LIMIT",  &wl_data);

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the QUEUE node
 * input:   bmf - our internal BMF message
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genQueueNode(char *queueName)
{
    xmlNodePtr queue_node = NULL;

    /* QUEUE node */ 
    queue_node = xmlNewNode(NULL, BAD_CAST "QUEUE");

    xmlNewChildString(queue_node, "NAME", queueName);

    // Prepare statistic data
    stat_data_t item_data, writer_data, reader_data;
    memset(&item_data,   0, sizeof(stat_data_t));
    memset(&writer_data, 0, sizeof(stat_data_t));
    memset(&reader_data, 0, sizeof(stat_data_t));

    /* ITEMS/WRITER/READER */
    item_data.current   = getItemsUsed(queueName);
    item_data.max       = getLoggedMaxItems(queueName);
    item_data.limit     = getItemsTotal(queueName);
    writer_data.current = getWriterCount(queueName);
    writer_data.max     = getLoggedMaxWriters(queueName);
    reader_data.current = getReaderCount(queueName);
    reader_data.max     = getLoggedMaxReaders(queueName);	

    xmlNewChildStat(queue_node, "ITEM",   &item_data);
    xmlNewChildStat(queue_node, "WRITER", &writer_data);
    xmlNewChildStat(queue_node, "READER", &reader_data);

    xmlAddChild(queue_node, genPacingNode(queueName));
	return queue_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the QUEUE_STATUS node
 * input:   bmf - our internal BMF message
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genQueueStatusNode(BMF bmf)
{
    xmlNodePtr node = xmlNewNode(NULL, BAD_CAST "QUEUE_STATUS");

    int count = 0; 
    count++; xmlAddChild(node, genQueueNode(PEER_QUEUE_NAME));
    count++; xmlAddChild(node, genQueueNode(LABEL_QUEUE_NAME));
    count++; xmlAddChild(node, genQueueNode(XML_U_QUEUE_NAME));
            
    xmlNewPropInt(node, "count", count);
    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: convert the chain status message to an XML text representation
 * input:   chainID - chain's ID
 * Output: the position of pointer in the buffer
 * He Yan @ Jun 22, 2008
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/ 
xmlNodePtr
genChainNode(int chainID)
{
    xmlNodePtr chain_node = NULL;
    /* CHAIN node */ chain_node = xmlNewNode(NULL, BAD_CAST "CHAIN");

	/* UPDATE LISTENER */
    /* ADDR/PORT */
	xmlNewChildString(chain_node, "UPDATE_ADDR", getChainAddress(chainID));
	xmlNewChildInt(chain_node,    "UPDATE_PORT", getUChainPort(chainID));

    /* STATE */
    xmlNewChildInt(chain_node, "UPDATE_STATE", getChainConnectionState(chainID, UPDATE_STREAM_CHAIN));
    
    /* COUNTs */
    time_data_t op_data;
    memset(&op_data,   0, sizeof(time_data_t));

    // Current value
	op_data.current   = getUChainConnectedTime(chainID);
	op_data.last_down = getULastConnectionDownTime(chainID);
    xmlNewChildTime(chain_node, "UPDATE_OPTIME",       &op_data);

    stat_data_t rv_data, rs_data;
    memset(&rv_data,   0, sizeof(stat_data_t));
    memset(&rs_data,   0, sizeof(stat_data_t));
    rv_data.current   = getUChainReceivedMessages(chainID);
    rs_data.current   = getUChainConnectionResetCount(chainID);
    xmlNewChildStat(chain_node, "UPDATE_RECV_MESSAGE", &rv_data);
    xmlNewChildStat(chain_node, "UPDATE_RESET",        &rs_data);

	return chain_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the CHAIN_STATUS node
 * input:   bmf - our internal BMF message
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genChainStatusNode(BMF bmf)
{
    xmlNodePtr node = xmlNewNode(NULL, BAD_CAST "CHAIN_STATUS");

    /*chains' status message*/
    int *chainIDs;
    int len = getActiveChainsIDs(&chainIDs);
    int count = 0; 
    int i;
    for(i=0; i<len; i++)
    {
        xmlAddChild(node, genChainNode(chainIDs[i]));
    }				
    /* update the count of prefixes */
    xmlNewPropInt(node, "count", count);
    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the SESSION node
 * input:   bmf - our internal BMF message
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genSessionNode(BMF bmf, int sessionID, int state_change)
{
    xmlNodePtr session_node = NULL;
    /* SESSION node */ session_node = xmlNewNode(NULL, BAD_CAST "SESSION");

    /* ADDR/PORT/AS */
    char *dstAddr = NULL;
    u_int16_t dstPort = 0;
    u_int16_t dstAS   = 0;
    Session_structp sp = getSessionByID(sessionID);
    if (sp)
    {
        dstAddr = sp->configInUse.remoteAddr;
        dstPort = sp->configInUse.remotePort;
        dstAS   = sp->configInUse.remoteAS2;
    }
	xmlNewChildString(session_node, "ADDR", dstAddr);
	xmlNewChildInt(session_node,    "PORT", dstPort);
	xmlNewChildInt(session_node,    "AS",   dstAS);

    if (state_change>0)
    {
        /* STATE_CHANGE */
        xmlAddChild(session_node, genBgpStateChangeNode(bmf));
    }
    else
    {
        /* STATE */
        xmlNewChildInt(session_node, "STATE", getSessionState(sessionID));
        
        /* COUNTs */
        time_data_t op_data;
        stat_data_t rv_data, rs_data;
        memset(&op_data,   0, sizeof(time_data_t));
        memset(&rv_data,   0, sizeof(stat_data_t));
        memset(&rs_data,   0, sizeof(stat_data_t));

        /* OPTIME/RECV_MESSAGE/RESET */
        op_data.current     = getSessionUPTime(sessionID);
        op_data.last_down   = getSessionLastDownTime(sessionID);
        op_data.last_action = getSessionLastActionTime(sessionID);

        rv_data.current     = getSessionMsgCount(sessionID);
        rs_data.current     = getSessionDownCount(sessionID);

        xmlNewChildTime(session_node, "OPTIME",       &op_data);
        xmlNewChildStat(session_node, "RECV_MESSAGE", &rv_data);
        xmlNewChildStat(session_node, "RESET",        &rs_data);

        /* COUNTs for session variables */
        stat_data_t pr_data,  // prefix
                    at_data,  // attirbute
                    mm_data,  // memory usage
                    na_data,  // announcements
                    da_data,  // dup announcements
                    sp_data,  // same path
                    dp_data,  // dup path
                    nw_data,  // withdrawal
                    dw_data;  // dup withdrawal
        memset(&pr_data,   0, sizeof(stat_data_t));
        memset(&at_data,   0, sizeof(stat_data_t));
        memset(&mm_data,   0, sizeof(stat_data_t));
        memset(&na_data,   0, sizeof(stat_data_t));
        memset(&da_data,   0, sizeof(stat_data_t));
        memset(&sp_data,   0, sizeof(stat_data_t));
        memset(&dp_data,   0, sizeof(stat_data_t));
        memset(&nw_data,   0, sizeof(stat_data_t));
        memset(&dw_data,   0, sizeof(stat_data_t));

        pr_data.current = getSessionPrefixCount(sessionID);
        at_data.current = getSessionAttributeCount(sessionID);
        mm_data.current = getSessionMemoryUsage(sessionID);

        na_data.current = getSessionCurrentNannCount(sessionID);
        da_data.current = getSessionCurrentDannCount(sessionID);
        sp_data.current = getSessionCurrentSPathCount(sessionID);
        dp_data.current = getSessionCurrentDPathCount(sessionID);
        nw_data.current = getSessionCurrentWithCount(sessionID);
        dw_data.current = getSessionCurrentDWithCount(sessionID);

        xmlNewChildStat(session_node, "PREFIX",       &pr_data);
        xmlNewChildStat(session_node, "ATTRIBUTE",    &at_data);
        xmlNewChildStat(session_node, "MEMORY_USAGE", &mm_data);

        xmlNewChildStat(session_node, "ANNOUNCEMENT",     &na_data);
        xmlNewChildStat(session_node, "DUP_ANNOUNCEMENT", &da_data);
        xmlNewChildStat(session_node, "SAME_PATH",        &sp_data);
        xmlNewChildStat(session_node, "DIFF_PATH",        &dp_data);
        xmlNewChildStat(session_node, "WITHDRAWAL",       &nw_data);
        xmlNewChildStat(session_node, "DUP_WITHDRAWAL",   &dw_data);


    }
	return session_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the SESSION_STATUS node
 * input:   bmf - our internal BMF message
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genSessionStatusNode(BMF bmf)
{
    xmlNodePtr node = xmlNewNode(NULL, BAD_CAST "SESSION_STATUS");

    int count = 0;
    int i = 0;
    switch ( bmf->type )
    {
        /* Periodic session status report */
        case BMF_TYPE_SESSION_STATUS:
        {
            for(i=0; i<MAX_SESSION_IDS; i++)
            {
                if( Sessions[i] != NULL )
                {
                    count++;
                    xmlAddChild(node, genSessionNode(bmf, i, 0));
                }
            }
            break;
        }
        /* Single session state change */
        case BMF_TYPE_FSM_STATE_CHANGE:
        {
            count++;
            xmlAddChild(node, genSessionNode(bmf, bmf->sessionID, 1));
            break;
        }
    }

    /* update the count of prefixes */
    xmlNewPropInt(node, "count", count);

    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the bgpmon node
 * input:   bmf - our internal BMF message
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpmonNode()
{
    xmlNodePtr node = xmlNewNode(NULL, BAD_CAST "BGPMON");

    char *addr;
    u_int16_t port;

    /* BGPMON NODE */
    // Client listen address and port
    addr = ClientControls.listenUAddr;
    port = ClientControls.listenUPort;
    /* ADDR */ xmlNewChildString(node, "ADDR", addr);
    /* PORT */ xmlNewChildInt(node,    "PORT", port);

    return node;
}


/*----------------------------------------------------------------------------------------
 * Purpose: generate BGPMON_START node
 * input:   bmf - pointer to BMF structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpmonStartNode(BMF bmf)
{
    /* BGPMON_START node */
    xmlNodePtr bgpmon_node = NULL;
    bgpmon_node = xmlNewNode(NULL, BAD_CAST "START");
    return bgpmon_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate BGPMON_STOP node
 * input:   bmf - pointer to BMF structure
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * He Yan @ Jun 22, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpmonStopNode(BMF bmf)
{
    /* BGPMON_STOP node */
    xmlNodePtr bgpmon_node = NULL;
    bgpmon_node = xmlNewNode(NULL, BAD_CAST "STOP");
    return bgpmon_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the BGPMON_STATUS node
 * input:   bmf - our internal BMF message
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpmonStatusNode(BMF bmf)
{
    xmlNodePtr node = xmlNewNode(NULL, BAD_CAST "BGPMON_STATUS");

    switch ( bmf->type )
    {
        case BMF_TYPE_BGPMON_START: { /* bgpmon start */ xmlAddChild(node, genBgpmonStartNode(bmf)); break; }
        case BMF_TYPE_BGPMON_STOP:  { /* bgpmon stop  */ xmlAddChild(node, genBgpmonStopNode(bmf));  break; }
    }
    return node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate the STATUS_MSG node
 * input:   bmf - our internal BMF message
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genStatusMsgNode(BMF bmf)
{
    xmlNodePtr status_node = NULL;

    /*
     * Creates BGP_MESSAGE node
     */
    status_node = xmlNewNode(NULL, BAD_CAST "STATUS_MSG");
    xmlAddChild(status_node, genBgpmonNode());

    switch ( bmf->type )
    {
        /* bgpmon status  */  
        case BMF_TYPE_BGPMON_START:
        case BMF_TYPE_BGPMON_STOP:
        {
            xmlAddChild(status_node, genBgpmonStatusNode(bmf));
            break;
        }
        /* queue status */
		case BMF_TYPE_QUEUES_STATUS:
        {
            xmlAddChild(status_node, genQueueStatusNode(bmf));
            break;
        }
        /* chain status */
		case BMF_TYPE_CHAINS_STATUS:
        {
            xmlAddChild(status_node, genChainStatusNode(bmf));
            break;
        }
        /* session state change */
        case BMF_TYPE_FSM_STATE_CHANGE:
        {
            if ( getSessionByID(bmf->sessionID) == NULL ) return 0;
            xmlAddChild(status_node, genSessionStatusNode(bmf));
            break;
        }
        /* session status */
        case BMF_TYPE_SESSION_STATUS:
        {
            xmlAddChild(status_node, genSessionStatusNode(bmf));
            break;
        }
    }

    return status_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate BGP_MESSAGE node
 * input:   bmf - our internal BMF message
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpMessageNodeWithDummyLength(BMF bmf)
{
    xmlNodePtr bgp_message_node = NULL; /* node pointers */

    /*
    * Creates BGP_MESSAGE node
    */
    bgp_message_node = xmlNewNode(NULL, BAD_CAST "BGP_MESSAGE");

    int xml_len = XML_BUFFER_LEN;
    char *type_str;

    /* Child nodes */
    switch ( bmf->type )
    {
        case BMF_TYPE_MSG_TO_PEER:
        case BMF_TYPE_MSG_FROM_PEER:
        case BMF_TYPE_MSG_LABELED:
        {
            /* time          */ xmlAddChild(bgp_message_node, genTimeNode(bmf));
            /* peering       */ xmlAddChild(bgp_message_node, genPeeringNode(bmf));
            /* ascii message */ xmlAddChild(bgp_message_node, genAsciiMsgNode(bmf));
            /* octet message */ xmlAddChild(bgp_message_node, genOctetMsgNode(bmf));
            type_str = "MESSAGE";
            break;
        }
        case BMF_TYPE_TABLE_TRANSFER:
        {
            /* time          */ xmlAddChild(bgp_message_node, genTimeNode(bmf));
            /* peering       */ xmlAddChild(bgp_message_node, genPeeringNode(bmf));
            /* pseudo message*/ xmlAddChild(bgp_message_node, genAsciiMsgNode(bmf));
            type_str = "TABLE";
            break;
            break;
        }
        /* State change messages */
        case BMF_TYPE_FSM_STATE_CHANGE:
        {
            /* time           */ xmlAddChild(bgp_message_node, genTimeNode(bmf));
            /* peering        */ xmlAddChild(bgp_message_node, genPeeringNode(bmf));
            /* status message */ xmlAddChild(bgp_message_node, genStatusMsgNode(bmf));
            type_str = "STATUS";
            break;
        }
        /* Status messages */
		case BMF_TYPE_CHAINS_STATUS:
		case BMF_TYPE_QUEUES_STATUS:
		case BMF_TYPE_SESSION_STATUS:
        case BMF_TYPE_BGPMON_START:
        case BMF_TYPE_BGPMON_STOP:
        {
            /* time           */ xmlAddChild(bgp_message_node, genTimeNode(bmf));
            /* status message */ xmlAddChild(bgp_message_node, genStatusMsgNode(bmf));
            type_str = "STATUS";
            break;
        }
        default:
        {
            log_err ("BMF2XML: unknown type!");
            type_str = "UNKNOWN";
            break;
        }
    }

    /* Attributes (Properties) */
    xmlNewPropInt(bgp_message_node,    "length",  xml_len);  /* dummy length with the maximum length */
    xmlNewPropString(bgp_message_node, "version", _VERSION); /* XFB version */
    xmlNewPropString(bgp_message_node, "xmlns",   _XMLNS);   /* XFB namespace */
    xmlNewPropInt(bgp_message_node,    "type_value",  bmf->type); /* bgpmon message type value */
    xmlNewPropString(bgp_message_node, "type",        type_str);  /* bgpmon message type */

    return bgp_message_node;
}

/*----------------------------------------------------------------------------------------
 * Purpose: generate BGP_MESSAGE node
 * input:   bmf - our internal BMF message
 * Output:  the new xml node
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
xmlNodePtr
genBgpMessageNodeWithStr(BMF bmf, char* xml)
{
    xmlNodePtr bgp_message_node = NULL; /* node pointers */

    /*-------------------------------------------------------
     * Create a BGP_MESSAGE node with dummy length attribute
     *------------------------------------------------------*/
    bgp_message_node = genBgpMessageNodeWithDummyLength(bmf);

    /*-------------------------------------------------------
     * Steps to replace the dummy length attribute
     * . Dump XML to a string buffer
     * . Obtain the string length
     * . Replace dummy length with the real length value
     *------------------------------------------------------*/
    xmlBufferPtr buff = xmlBufferCreate();
    xmlNodeDump(buff, NULL, bgp_message_node, 0, 0);

    static char format[XML_TEMP_BUFFER_LEN];
    static char length_str[XML_TEMP_BUFFER_LEN];
    static char old_length_str[XML_TEMP_BUFFER_LEN];
    static char new_length_str[XML_TEMP_BUFFER_LEN];
    format[0] = '\0';

    /* Generate format */
    if (_XML_LEN_DIGITS == 0) 
        _XML_LEN_DIGITS = ceil(log10(XML_BUFFER_LEN));

    sprintf(format, "%s%d.%dd", "%", _XML_LEN_DIGITS, _XML_LEN_DIGITS);
    /* Genearte length string from format and REAL length */
    sprintf(length_str, format, strlen((char *)buff->content));
    strcpy(xml,  (char *)buff->content);

    /* Replace length string */
    sprintf(old_length_str, "length=\"%d\"", XML_BUFFER_LEN); /* Old length string */
    sprintf(new_length_str, "length=\"%s\"", length_str);     /* New length string */
    replace_str(xml, old_length_str, new_length_str);

    /* Update length attribute */
    xmlSetProp (bgp_message_node, BAD_CAST "length", BAD_CAST length_str);
    xmlBufferFree(buff);  /* free xml buffer       */

    return bgp_message_node;
}


/*----------------------------------------------------------------------------------------
 * Purpose: entry fucntion which converts all types of BMF messages to XML text representations
 * input:   bmf - our internal BMF message
 *          xml - pointer to the buffer used to store the result XML string
 *          maxlen - max length of the buffer
 * Pei-chun Cheng @ Dec 20, 2008
 * -------------------------------------------------------------------------------------*/
int BMF2XMLDATA(BMF bmf, char *xml, int maxlen)
{
    xmlInitParser();

    xmlDocPtr  doc              = NULL;  /* document pointer */
    xmlNodePtr bgp_message_node = NULL;  /* xml node pointer */

    /*-------------------------------------------------------
     * Implementation:
     *------------------------------------------------------*/
    doc = xmlNewDoc(BAD_CAST "1.0");
    bgp_message_node = genBgpMessageNodeWithStr(bmf, xml);
    xmlDocSetRootElement(doc, bgp_message_node);
    strcpy(_XML, xml);

    /*-------------------------------------------------------
     * Clean up
     *------------------------------------------------------*/
    xmlFreeDoc(doc);      /* Free the xml document */

    return strlen(xml);
}


/* vim: sw=4 ts=4 sts=4 expandtab
 */
